import os
import shutil

import numpy as np
import pytest
import torch
import torch.nn as nn

import fannypack


class SimpleNet(nn.Module):
    """Simple PyTorch model. Scalar input, scalar output."""

    def __init__(self):
        super().__init__()

        # Define layers
        self.layer1 = nn.Linear(1, 8)
        self.layer2 = nn.Linear(8, 8)
        self.layer3 = nn.Linear(8, 8)
        self.layer4 = nn.Linear(8, 1)

        # Compose layers w/ nonlinearities
        self.layers = nn.Sequential(
            self.layer1,
            nn.ReLU(inplace=True),
            self.layer2,
            nn.ReLU(inplace=True),
            self.layer3,
            nn.ReLU(inplace=True),
            self.layer4,
        )

    def forward(self, x):
        return self.layers(x)


class ResBlockNet(nn.Module):
    """Simple PyTorch model. Scalar input, scalar output."""

    def __init__(self):
        super().__init__()

        # Define layers
        self.layer1 = nn.Linear(1, 9)
        self.layer2 = fannypack.nn.resblocks.Linear(9)
        self.layer3 = fannypack.nn.resblocks.Conv2d(1)
        self.layer4 = fannypack.nn.resblocks.Linear(9)
        self.layer5 = nn.Linear(9, 1)

    def forward(self, x):
        x = self.layer1(x)
        x = self.layer2(x).view(-1, 1, 3, 3)
        x = self.layer3(x).view(-1, 9)
        x = self.layer4(x)
        x = self.layer5(x)
        return x


@pytest.fixture
def simple_net():
    """Constructs an MLP implemented in PyTorch."""
    # Deterministic tests are nice..
    torch.manual_seed(0)
    return SimpleNet()


@pytest.fixture()
def simple_buddy():
    """Fixture for setting up a Buddy, as well as some dummy training data."""
    # Deterministic tests are nice..
    np.random.seed(0)
    torch.manual_seed(0)

    # Construct neural net, training buddy
    simple_net = SimpleNet()
    buddy = fannypack.utils.Buddy(
        "simple_net",
        simple_net,
        # Use directories relative to this fixture
        checkpoint_dir=os.path.join(os.path.dirname(__file__), "assets/checkpoints/"),
        metadata_dir=os.path.join(os.path.dirname(__file__), "assets/metadata/"),
        log_dir=os.path.join(os.path.dirname(__file__), "assets/logs/"),
        verbose=True,
        cpu_only=True,
    )

    # Batch size
    N = 20

    # Learn to regress a constant
    data = torch.FloatTensor(np.random.normal(size=(N, 1)))
    labels = torch.FloatTensor(np.random.normal(loc=3, size=(1, 1))).expand((N, 1))
    return simple_net, buddy, data, labels


@pytest.fixture()
def simple_buddy_temporary_data():
    """Fixture for setting up a Buddy, as well as some dummy training data."""
    # Deterministic tests are nice..
    np.random.seed(0)
    torch.manual_seed(0)

    # Construct neural net, training buddy
    simple_net = SimpleNet()
    buddy = fannypack.utils.Buddy(
        "temporary_net",
        simple_net,
        # Use directories relative to this fixture
        checkpoint_dir=os.path.join(
            os.path.dirname(__file__), "tmp/assets/checkpoints/"
        ),
        metadata_dir=os.path.join(os.path.dirname(__file__), "tmp/assets/metadata/"),
        log_dir=os.path.join(os.path.dirname(__file__), "tmp/assets/logs/"),
        verbose=True,
        # Disable auto-checkpointing
        optimizer_checkpoint_interval=0,
        cpu_only=True,
    )

    # Batch size
    N = 20

    # Learn to regress a constant
    data = torch.FloatTensor(np.random.normal(size=(N, 1)))
    labels = torch.FloatTensor(np.random.normal(loc=3, size=(1, 1))).expand((N, 1))
    yield simple_net, buddy, data, labels

    # Delete temporary files when done
    path = os.path.join(os.path.dirname(__file__), "tmp/")
    if os.path.isdir(path):
        shutil.rmtree(path)


@pytest.fixture()
def resblock_buddy_temporary_data():
    """Fixture for setting up a Buddy, as well as some dummy training data.

    This is similar to `simple_buddy`, but uses temporary directories for
    checkpointing, metadata saving, and logging. Saved files are deleted
    automatically when the test exits.
    """
    # Deterministic tests are nice..
    np.random.seed(0)
    torch.manual_seed(0)

    # Construct neural net, training buddy
    resblock_net = ResBlockNet()
    buddy = fannypack.utils.Buddy(
        "resblock_net",
        resblock_net,
        # Use directories relative to this fixture
        checkpoint_dir=os.path.join(
            os.path.dirname(__file__), "tmp/assets/checkpoints/"
        ),
        metadata_dir=os.path.join(os.path.dirname(__file__), "tmp/assets/metadata/"),
        log_dir=os.path.join(os.path.dirname(__file__), "tmp/assets/logs/"),
        verbose=True,
        cpu_only=True,
    )

    # Batch size
    N = 20

    # Learn to regress a constant
    data = torch.FloatTensor(np.random.normal(size=(N, 1)))
    labels = torch.FloatTensor(np.random.normal(loc=3, size=(1, 1))).expand((N, 1))
    yield resblock_net, buddy, data, labels

    # Delete temporary files when done
    path = os.path.join(os.path.dirname(__file__), "tmp/")
    if os.path.isdir(path):
        shutil.rmtree(path)



import abc
import pathlib
from typing import TYPE_CHECKING, Any, Dict, cast

import yaml

if TYPE_CHECKING:
    from .._buddy import Buddy


class _BuddyMetadata(abc.ABC):
    """Buddy's experiment metadata management interface."""

    def __init__(self, metadata_dir: str) -> None:
        """Metadata-specific setup."""
        self._metadata_dir = metadata_dir
        try:
            # Attempt to read existing metadata
            self.load_metadata(_write=True)
        except FileNotFoundError:
            self._metadata: Dict[str, Any] = {}

    def load_metadata(
        self,
        experiment_name: str = None,
        metadata_dir: str = None,
        path: str = None,
        _write=True,
    ) -> None:
        """Read existing metadata file. Note that metadata is loaded automatically: this
        only needs to be called if loading across experiments.

        Overwrites existing metadata.
        """
        if path is None:
            if experiment_name is None:
                experiment_name = cast("Buddy", self)._experiment_name
            if metadata_dir is None:
                metadata_dir = self._metadata_dir
            path = str(pathlib.Path(metadata_dir) / f"{experiment_name}.yaml")
        else:
            assert experiment_name is None and metadata_dir is None

        # Read from disk
        with open(path, "r") as file:
            self._metadata = yaml.load(file, Loader=yaml.Loader)
            cast("Buddy", self)._print("Loaded metadata:", self._metadata)

        # Write to disk
        if _write:
            self._write_metadata()

    def add_metadata(self, content: Dict[str, Any]) -> None:
        """Add human-readable metadata for this experiment. Input should be a
        dictionary that is merged with existing metadata.
        """
        assert type(content) == dict

        # Merge input metadata with current metadata
        for key, value in content.items():
            self._metadata[key] = value

        # Write to disk
        self._write_metadata()

    def set_metadata(self, content: Dict[str, Any]) -> None:
        """Assign human-readable metadata for this experiment. Input should be
        a dictionary that replaces existing metadata.
        """
        assert type(content) == dict

        # Replace input metadata with current metadata
        self._metadata = content

        # Write to disk
        self._write_metadata()

    def _write_metadata(self) -> None:
        # Create metadata directory if needed
        metadata_dir = pathlib.Path(self._metadata_dir)
        assert not metadata_dir.is_file()
        if not metadata_dir.exists():
            metadata_dir.mkdir(parents=True)
            cast("Buddy", self)._print("Created directory:", metadata_dir)

        # Write metadata to file
        metadata_path = metadata_dir / f"{cast('Buddy', self)._experiment_name}.yaml"
        with open(metadata_path, "w") as file:
            yaml.dump(self._metadata, file, default_flow_style=False)
            cast("Buddy", self)._print("Wrote metadata to:", metadata_path)

    @property
    def metadata(self) -> Dict[str, Any]:
        """Read-only interface for experiment metadata."""
        return self._metadata

    @property
    def metadata_path(self) -> str:
        """Read-only path to my metadata file."""
        return str(
            pathlib.Path(self._metadata_dir)
            / f"{cast('Buddy', self)._experiment_name}.yaml"
        )



import pdb
import signal
import sys
import traceback as tb


def pdb_safety_net():
    """Attaches a "safety net" for unexpected errors in a Python script.

    When called, PDB will be automatically opened when either (a) the user hits Ctrl+C
    or (b) we encounter an uncaught exception. Helpful for bypassing minor errors,
    diagnosing problems, and rescuing unsaved models.
    """

    # Open PDB on Ctrl+C
    def handler(sig, frame):
        pdb.set_trace()

    signal.signal(signal.SIGINT, handler)

    # Open PDB when we encounter an uncaught exception
    def excepthook(type_, value, traceback):  # pragma: no cover (impossible to test)
        tb.print_exception(type_, value, traceback, limit=100)
        pdb.post_mortem(traceback)

    sys.excepthook = excepthook



import glob
import os
from dataclasses import dataclass
from typing import Dict, List, Optional, Set, cast


def _listdir(path: str) -> List[str]:
    """Helper for listing files in a directory"""
    try:
        return os.listdir(path)
    except FileNotFoundError:
        return []


@dataclass(frozen=True)
class FindOutput:
    """Output of `find_experiments(...)`."""

    experiment_names: Set[str]
    checkpoint_counts: Dict[str, int]
    log_experiments: Set[str]
    metadata_experiments: Set[str]
    timestamps: Dict[str, float]


def get_size(path: str) -> int:
    """Returns the size in bytes of the file or directory located at a path."""
    if os.path.isfile(path):
        return os.stat(path).st_size
    elif os.path.isdir(path):
        return sum(
            [
                os.stat(p).st_size
                for p in glob.glob(os.path.join(path, "**/*"), recursive=True)
            ]
        )
    else:
        return 0


def format_size(size: float, short: bool = False):
    """Converts a size in bytes to a human-readable string."""
    units = ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]
    unit_index = 0
    while size > 1024 and unit_index < len(units) - 1:
        size /= 1024.0
        unit_index += 1

    if short:
        units = ["B", "K", "M", "G", "T", "P"]
        return f"{size:.0f}{units[unit_index]}"
    else:
        return f"{size:.2f} {units[unit_index]}"


@dataclass
class BuddyPaths:
    """Dataclass for storing paths to experiment files."""

    checkpoint_dir: str
    log_dir: str
    metadata_dir: str

    _find_output_cache: Optional[FindOutput] = None

    def find_checkpoints(self, experiment_name: str):
        """Finds checkpoints associated with an experiment."""
        # Glob for checkpoint files
        checkpoint_files = glob.glob(
            os.path.join(self.checkpoint_dir, f"{glob.escape(experiment_name)}-*.ckpt")
        )

        # Filter further with rpartition (for handling experiment names with hyphens)
        return list(
            filter(
                lambda path: path.rpartition("-")[0].endswith(experiment_name),
                checkpoint_files,
            )
        )

    def get_log_dir(self, experiment_name: str):
        """Returns a path to an experiment's log directory."""
        return os.path.join(self.log_dir, experiment_name)

    def get_metadata_file(self, experiment_name: str):
        """Returns a path to an experiment's metadata file."""
        return os.path.join(self.metadata_dir, f"{experiment_name}.yaml")

    def find_experiments(self, verbose: bool = False) -> FindOutput:
        """Helper for listing experiments"""

        # Return cached results
        if self._find_output_cache is not None:
            return cast(FindOutput, self._find_output_cache)

        # Print helper
        def _print(*args, **kwargs):
            if not verbose:
                return
            print(*args, **kwargs)

        # Last modified: checkpoints and metadata only
        # > We could also do logs, but seems high effort?
        timestamps: Dict[str, float] = {}

        # Count checkpoints for each experiment
        checkpoint_counts: Dict[str, int] = {}
        for file in _listdir(self.checkpoint_dir):
            # Remove .ckpt suffix
            if file[-5:] != ".ckpt":
                _print(f"Skipping malformed checkpoint filename: {file}")
                continue
            trimmed = file[:-5]

            # Get experiment name
            name, hyphen, _label = trimmed.rpartition("-")
            if hyphen != "-":
                _print(f"Skipping malformed checkpoint filename: {file}")
                continue

            # Update tracker
            if name not in checkpoint_counts.keys():
                checkpoint_counts[name] = 0
            checkpoint_counts[name] += 1

            # Update timestamp
            mtime = os.path.getmtime(os.path.join(self.checkpoint_dir, file))
            if name not in timestamps.keys() or mtime > timestamps[name]:
                timestamps[name] = mtime

        # Get experiment names from metadata files
        metadata_experiments = set()
        for file in _listdir(self.metadata_dir):
            # Remove .yaml suffix
            if file[-5:] != ".yaml":
                _print(f"Skipping malformed metadata filename: {file}")
                continue
            name = file[:-5]
            metadata_experiments.add(name)

            # Update timestamp
            mtime = os.path.getmtime(os.path.join(self.metadata_dir, file))
            if name not in timestamps.keys() or mtime > timestamps[name]:
                timestamps[name] = mtime

        # Get experiment names from log directories
        log_experiments = set()
        for experiment_name in _listdir(self.log_dir):
            path = os.path.join(self.log_dir, experiment_name)
            if len(_listdir(path)) > 0:
                # Log files exist
                log_experiments.add(experiment_name)
            else:
                # Clean up empty directory
                os.rmdir(path)

        # Get all experiments
        experiment_names = (
            set(checkpoint_counts.keys()) | log_experiments | metadata_experiments
        )

        # Update global find_output cache
        self._find_output_cache = FindOutput(
            experiment_names=experiment_names,
            checkpoint_counts=checkpoint_counts,
            log_experiments=log_experiments,
            metadata_experiments=metadata_experiments,
            timestamps=timestamps,
        )
        return self._find_output_cache

    def clear_cache(self):
        """Clears the experiment list cache."""
        self._find_output_cache = None



import os
import subprocess


def get_git_commit_hash(path: str = "./") -> str:
    """Returns the current Git commit hash.

    Args:
        path (str, optional): Path to check. Defaults to './'.
    """
    return (
        subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=os.path.dirname(path))
        .decode("ascii")
        .strip()
    )



# PYTHON_ARGCOMPLETE_OK

# To print options, install `fannypack` via pip and run `$ buddy --help`

"""CLI interface for experiment management via Buddy.

By default, searches for checkpoints in `./checkpoints/`, logs in `./logs/`, and
metadata in `./metadata/`. To override these values, set the BUDDY_CHECKPOINT_DIR,
BUDDY_LOG_DIR, and BUDDY_METADATA_DIR environment variables.
"""
import argparse
import os
from typing import Dict, Type

import argcomplete

from ._buddy_cli_include import BuddyPaths, Subcommand, subcommand_registry


def main() -> None:
    # Pull paths from environment
    paths = BuddyPaths(
        checkpoint_dir=os.environ.get("BUDDY_CHECKPOINT_DIR", default="checkpoints/"),
        log_dir=os.environ.get("BUDDY_LOG_DIR", default="logs/"),
        metadata_dir=os.environ.get("BUDDY_METADATA_DIR", default="metadata/"),
    )

    # Set up argument parser
    parser = argparse.ArgumentParser(
        prog="buddy",
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    # Separate parsers for subcommands
    subparsers = parser.add_subparsers(
        # required=True, <= Not supported in Python 3.6
        dest="subcommand",
        help="Get help by running `$ buddy {subcommand} --help`.",
    )

    # Add subcommands
    subcommand_map: Dict[str, Type[Subcommand]] = {}
    for S in subcommand_registry:
        subparser = subparsers.add_parser(
            S.subcommand, help=S.__doc__, description=S.__doc__
        )
        S.add_arguments(parser=subparser, paths=paths)
        subcommand_map[S.subcommand] = S

    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    # Run subcommand
    if args.subcommand is None:
        print(parser.format_help())
        exit(2)
    else:
        subcommand_map[args.subcommand].main(args=args, paths=paths)


if __name__ == "__main__":
    main()



import numpy as np
import pytest

import fannypack


def test_squeeze_simple():
    X = np.zeros((10, 1, 5, 1, 10))
    X = fannypack.utils.squeeze(X)
    assert X.shape == (10, 5, 10)


def test_squeeze_axis():
    X = np.zeros((10, 1, 5, 1, 10))
    X = fannypack.utils.squeeze(X, axis=1)
    assert X.shape == (10, 5, 1, 10)


def test_squeeze_bad_axis():
    X = np.zeros((10, 1, 5, 1, 10))
    with pytest.raises(AssertionError):
        X = fannypack.utils.squeeze(X, axis=0)



# Register subcommands
from . import _subcommand_delete, _subcommand_info, _subcommand_list, _subcommand_rename

# Expose important stuff for CLI script
from ._subcommand import Subcommand, subcommand_registry
from ._utils import BuddyPaths

__all__ = [
    "Subcommand",
    "subcommand_registry",
    "BuddyPaths",
]



import abc
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Union, cast

import torch

if TYPE_CHECKING:
    from .._buddy import Buddy
    from ._checkpointing import _BuddyCheckpointing


@dataclass
class _BuddyOptimizerConfig:
    """Optimizer configuration.

    Along with the state dict of each optimizer, this is saved with each checkpoint.
    """

    global_steps: int
    """int: Total number of optimization steps taken. This is used for logging, LR schedulers, etc."""

    optimizer_type: str
    """str: Type of optimizer to used. Currently only `adam` or `adadelta`."""

    learning_rate_schedulers: Dict[str, Callable[[int], float]]
    """dict: Dictionary mapping optimizer names to LR scheduling functions."""


class _BuddyOptimizer(abc.ABC):
    """Buddy's optimization interface."""

    # Supported optimizer types
    # TODO: improve typing here; Type[torch.optim.Optimizer] would work but the
    # superclass constructor doesn't match the subclass constructors
    _OPTIMIZER_TYPES: Dict[str, Any] = {
        "adam": torch.optim.Adam,
        "adadelta": torch.optim.Adadelta,
    }

    def __init__(
        self, optimizer_type: str, optimizer_checkpoint_interval: float
    ) -> None:
        """Optimizer-specific setup."""
        # Assign our training configuration.
        self._optimizer_config = _BuddyOptimizerConfig(
            global_steps=0,
            optimizer_type=optimizer_type,
            learning_rate_schedulers={},
        )

        # Map from optimizer name to optimizers
        # These are constructed lazily!
        self._optimizer_dict: Dict[str, torch.optim.Optimizer] = {}

        # Autocheckpoint variables
        self._optimizer_checkpoint_interval: float = optimizer_checkpoint_interval
        self._optimizer_last_checkpoint_time: Optional[float] = None

        # Default learning rate
        self._optimizer_default_learning_rate: Optional[
            Union[float, Callable[[int], float]]
        ] = None

    def minimize(
        self,
        loss: torch.Tensor,
        optimizer_name: str = "primary",
        *,
        retain_graph: bool = False,
        checkpoint_interval: Optional[float] = None,
        clip_grad_max_norm: Optional[float] = None,
    ) -> None:
        """Compute gradients and use them to minimize a loss function."""
        model = cast("Buddy", self)._model
        assert model is not None, "No model attached!"

        # Get optimizer
        self._instantiate_optimizer(optimizer_name)
        optimizer: torch.optim.Optimizer = self._optimizer_dict[optimizer_name]

        # Update learning rate using scheduler if possible
        schedulers = self._optimizer_config.learning_rate_schedulers
        if optimizer_name in schedulers:
            self._set_learning_rate(
                schedulers[optimizer_name](self._optimizer_config.global_steps),
                optimizer_name,
            )

        # Take gradient step
        optimizer.zero_grad()
        loss.backward(retain_graph=retain_graph)  # type: ignore
        if clip_grad_max_norm is not None:
            torch.nn.utils.clip_grad_norm_(
                optimizer.param_groups[0]["params"],
                max_norm=clip_grad_max_norm,
            )
        optimizer.step()

        # Update global step count
        self._optimizer_config.global_steps += 1

        # Autocheckpoint procedure
        if checkpoint_interval is None:
            checkpoint_interval = self._optimizer_checkpoint_interval

        # Disable autocheckpoint if interval is 0
        if checkpoint_interval == 0:
            return

        if self._optimizer_last_checkpoint_time is None:
            # First iteration
            self._optimizer_last_checkpoint_time = time.time()
        elif (
            time.time() - cast(float, self._optimizer_last_checkpoint_time)
            > self._optimizer_checkpoint_interval
        ):  # pragma: no cover
            # Checkpoint!
            cast("_BuddyCheckpointing", self).save_checkpoint()
            self._optimizer_last_checkpoint_time = time.time()

    def get_learning_rate(self, optimizer_name: str = "primary") -> float:
        """Gets an optimizer learning rate."""
        assert cast("Buddy", self)._model is not None, "No model attached!"
        assert optimizer_name in self._optimizer_dict

        # Return scheduled learning rate
        schedulers = self._optimizer_config.learning_rate_schedulers
        if optimizer_name in schedulers:
            return schedulers[optimizer_name](self.optimizer_steps)

        # Return raw learning rate
        # Currently, only one parameter group is supported
        optimizer = self._optimizer_dict[optimizer_name]
        assert len(optimizer.param_groups) == 1
        return optimizer.param_groups[0]["lr"]

    def set_learning_rate(
        self,
        value: Union[float, Callable[[int], float]],
        optimizer_name: str = "primary",
    ) -> None:
        """Sets an optimizer learning rate. Accepts either a floating point
        learning rate or a schedule function (int steps -> float LR).
        """
        assert cast("Buddy", self)._model is not None, "No model attached!"
        schedulers = self._optimizer_config.learning_rate_schedulers
        if callable(value):
            assert isinstance(value(0), float)

            # Make sure optimizer is instantiated: if not, learning rate will be
            # overriden when it is
            self._instantiate_optimizer(optimizer_name)

            # Store scheduler
            schedulers[optimizer_name] = value
        else:
            # Set learning rate to a float
            assert isinstance(value, float)
            # Delete scheduler
            if optimizer_name in schedulers.keys():
                schedulers.pop(optimizer_name)

            # Set scalar learning rate
            self._set_learning_rate(value, optimizer_name)

    def set_default_learning_rate(
        self, value: Union[float, Callable[[int], float]]
    ) -> None:
        """Sets a default learning rate for new optimizers."""
        self._optimizer_default_learning_rate = value

    @property
    def optimizer_steps(self) -> int:
        """Read-only interface for # of steps taken by optimizer."""
        return self._optimizer_config.global_steps

    def _set_learning_rate(self, value: float, optimizer_name: str) -> None:
        """(Private) Sets an optimizer's learning rate."""

        self._instantiate_optimizer(optimizer_name)

        # Currently, only one parameter group is supported
        optimizer = self._optimizer_dict[optimizer_name]
        assert len(optimizer.param_groups) == 1
        optimizer.param_groups[0]["lr"] = value

    def _instantiate_optimizer(self, optimizer_name: str) -> None:
        """(Private) Instantiates an optimizer. Returns immediately if
        optimizer already exists.
        """
        assert cast("Buddy", self)._model is not None, "No model attached!"
        if optimizer_name in self._optimizer_dict.keys():
            # Optimizer already exists: do nothing!
            return

        cast("Buddy", self)._print("Instantiating optimizer: ", optimizer_name)

        # Make sure we're creating a valid optimizer
        optimizer_type = self._optimizer_config.optimizer_type
        assert optimizer_type in self._OPTIMIZER_TYPES

        # Parameters
        optimizer_class = self._OPTIMIZER_TYPES[optimizer_type]

        # Construct optimizer
        self._optimizer_dict[optimizer_name] = optimizer_class(
            cast("Buddy", self).model.parameters()
        )

        # Set default learning rate
        if self._optimizer_default_learning_rate is not None:
            self.set_learning_rate(
                self._optimizer_default_learning_rate, optimizer_name=optimizer_name
            )



import abc
from typing import Callable, Dict

import torch
import torch.nn as nn


class Base(nn.Module, abc.ABC):
    """Base class for a generic residual block, with support for `"relu"`,
    `"leaky_relu"`, `"selu"`, and `"none"` activations.
    """

    _activation_types: Dict[str, Callable[..., nn.Module]] = {
        "relu": nn.ReLU,
        "leaky_relu": nn.LeakyReLU,
        "selu": nn.SELU,
        "none": nn.Identity,
    }
    """Dictionary of support activation types.
    """

    def __init__(self, activation: str = "relu", activations_inplace: bool = False):
        super().__init__()

        self.block1: nn.Module
        self.block2: nn.Module
        self.activation: nn.Module = self._activation_types[activation](
            inplace=activations_inplace
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        """ResBlock forward pass."""
        residual = x
        x = self.block1(x)
        x = self.activation(x)
        x = self.block2(x)
        assert x.shape[0] == residual.shape[0]
        x += residual
        x = self.activation(x)
        return x


class Linear(Base):
    """Standard linear residual block."""

    def __init__(self, units: int, bottleneck_units: int = None, **resblock_base_args):
        super().__init__(**resblock_base_args)

        if bottleneck_units is None:
            bottleneck_units = units
        self.block1 = nn.Linear(units, bottleneck_units)
        self.block2 = nn.Linear(bottleneck_units, units)


class Conv2d(Base):
    """Standard convolutional residual block."""

    def __init__(
        self,
        channels: int,
        bottleneck_channels: int = None,
        kernel_size: int = 3,
        **resblock_base_args
    ):
        super().__init__(**resblock_base_args)

        if bottleneck_channels is None:
            bottleneck_channels = channels

        self.block1 = nn.Conv2d(
            channels,
            bottleneck_channels,
            kernel_size=kernel_size,
            padding=kernel_size // 2,
        )
        self.block2 = nn.Conv2d(
            bottleneck_channels,
            channels,
            kernel_size=kernel_size,
            padding=kernel_size // 2,
        )


__all__ = ["Base", "Linear", "Conv2d"]



import numpy as np
import pytest

import fannypack


@pytest.fixture
def wrapper():
    """Creates a SliceWrapper around a dictionary"""
    raw = {"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}
    wrapper = fannypack.utils.SliceWrapper(raw)
    return wrapper


@pytest.fixture
def wrapper_complex():
    """Creates a SliceWrapper around a more complex dictionary"""
    raw = {
        "a": np.random.randint(10, size=(10, 5, 31)),
        "b": np.random.randint(10, size=(10, 5, 3)),
    }
    wrapper = fannypack.utils.SliceWrapper(raw)
    return wrapper


@pytest.fixture
def wrapper_thin():
    """Creates a SliceWrapper around a Python list."""
    raw = [1, 2, 3, 4]
    wrapper = fannypack.utils.SliceWrapper(raw)
    return wrapper


def test_map(wrapper):
    """Validate pre/post-conditions of a call to `map()`."""
    assert type(wrapper.data["a"]) == list
    wrapper = fannypack.utils.SliceWrapper(wrapper.map(np.asarray))
    assert type(wrapper.data["a"]) == np.ndarray


def test_thin_map(wrapper_thin):
    """Validate pre/post-conditions of a call to `map()`."""
    assert type(wrapper_thin.data) == list
    wrapper_thin = fannypack.utils.SliceWrapper(wrapper_thin.map(np.asarray))
    assert type(wrapper_thin.data) == np.ndarray


def test_append(wrapper):
    """Checks append interface"""
    new = {
        "a": 5,
        "b": 3,
    }
    wrapper.append(new)
    assert wrapper[-1] == new


def test_extend_new():
    """Checks extend interface"""
    wrapper = fannypack.utils.SliceWrapper({})
    new = {
        "a": [5],
        "b": [3],
    }
    wrapper.extend(new)
    assert wrapper[-1:] == new


def test_append_new():
    """Checks append interface"""
    wrapper = fannypack.utils.SliceWrapper({})
    new = {
        "a": 5,
        "b": 3,
    }
    wrapper.append(new)
    assert wrapper[-1] == new


def test_extend(wrapper):
    """Checks extend interface"""
    new = {
        "a": [5],
        "b": [3],
    }
    wrapper.extend(new)
    assert wrapper[-1:] == new


def test_append_thin(wrapper_thin):
    """Checks append interface (thin)"""
    new = 5
    wrapper_thin.append(new)
    assert wrapper_thin[-1] == new


def test_extend_thin(wrapper_thin):
    """Checks extend interface (thin)"""
    new = [5]
    wrapper_thin.extend(new)
    assert wrapper_thin[-1:] == new


def test_iterable(wrapper):
    """Check iterable interface."""
    counter = 0
    for x in wrapper:
        assert type(x) == dict
        assert type(x["a"]) == int
        counter += 1
    assert counter == len(wrapper)


def test_iterable_numpy(wrapper):
    """Check iterable interface. (numpy)"""
    wrapper = fannypack.utils.SliceWrapper(wrapper.map(np.asarray))
    counter = 0
    for x in wrapper:
        assert type(x) == dict
        assert type(x["a"]) == np.int64
        counter += 1
    assert counter == len(wrapper)


def test_iterable_thin(wrapper_thin):
    """Check iterable interface."""
    counter = 0
    for x in wrapper_thin:
        assert type(x) == int
        counter += 1
    assert counter == len(wrapper_thin)


def test_iterable_numpy_thin(wrapper_thin):
    """Check iterable interface. (numpy)"""
    wrapper_thin = fannypack.utils.SliceWrapper(wrapper_thin.map(np.asarray))
    counter = 0
    for x in wrapper_thin:
        assert type(x) == np.int64
        counter += 1
    assert counter == len(wrapper_thin)


def test_len(wrapper):
    """Check `len()` output."""
    assert len(wrapper) == 4
    wrapper = fannypack.utils.SliceWrapper(wrapper.map(np.asarray))
    assert len(wrapper) == 4


def test_shape(wrapper):
    """Check `shape` property."""
    assert wrapper.shape == (4,)
    wrapper = fannypack.utils.SliceWrapper(wrapper.map(np.asarray))
    assert wrapper.shape == (4,)


def test_shape_complex(wrapper_complex):
    """Check `shape` property."""
    assert wrapper_complex.shape == (10, 5)


def test_read_slice(wrapper):
    """Check that we can read slices of wrappers."""
    assert wrapper[::1]["a"] == [1, 2, 3, 4]
    assert wrapper[::1]["b"] == [5, 6, 7, 8]

    assert wrapper[::2]["a"] == [1, 3]
    assert wrapper[::2]["b"] == [5, 7]

    assert wrapper[:2:]["a"] == [1, 2]
    assert wrapper[1:3:]["b"] == [6, 7]


def test_read_slice_numpy(wrapper):
    """Check that we can read slices of our wrappers (numpy)."""
    wrapper = fannypack.utils.SliceWrapper(wrapper.map(np.asarray))

    assert np.allclose(wrapper[::1]["a"], [1, 2, 3, 4])
    assert np.allclose(wrapper[::1]["b"], [5, 6, 7, 8])

    assert np.allclose(wrapper[::2]["a"], [1, 3])
    assert np.allclose(wrapper[::2]["b"], [5, 7])

    assert np.allclose(wrapper[:2:]["a"], [1, 2])
    assert np.allclose(wrapper[1:3:]["b"], [6, 7])


def test_write_slice(wrapper):
    """Check that writing does nothing for wrappers containing raw lists."""
    wrapper[::1]["a"][0] = 1000
    test_read_slice(wrapper)


def test_write_slice_numpy(wrapper):
    """Check that we can write to slices for wrappers containing numpy arrays."""
    wrapper = fannypack.utils.SliceWrapper(wrapper.map(np.asarray))

    wrapper[::1]["a"][::2] = 0

    assert np.allclose(wrapper[::1]["a"], [0, 2, 0, 4])
    assert np.allclose(wrapper[::1]["b"], [5, 6, 7, 8])

    assert np.allclose(wrapper[::2]["a"], [0, 0])
    assert np.allclose(wrapper[::2]["b"], [5, 7])

    assert np.allclose(wrapper[:2:]["a"], [0, 2])
    assert np.allclose(wrapper[1:3:]["b"], [6, 7])



import numpy as np
import pytest
import torch
import torch.nn.functional as F

import fannypack

from ..fixtures import simple_buddy_temporary_data


def test_buddy_no_model():
    """Check that errors are raised if a Buddy is used without a model attached."""
    buddy = fannypack.utils.Buddy("no_model")

    with pytest.raises(AssertionError):
        buddy.model

    with pytest.raises(AssertionError):
        buddy.save_checkpoint()

    with pytest.raises(AssertionError):
        buddy.load_checkpoint()

    with pytest.raises(AssertionError):
        fake_loss = torch.Tensor([0.0]).to(buddy.device)
        buddy.minimize(fake_loss)


def test_buddy_log_scopes(simple_buddy_temporary_data):
    """Check that log scope functions as expected."""
    model, buddy, data, labels = simple_buddy_temporary_data

    assert buddy.log_scope_prefix() == ""
    with buddy.log_scope("scope0"):
        buddy.log_scope_push("scope1")
        with buddy.log_scope("scope2"):
            assert buddy.log_scope_prefix("name") == "scope0/scope1/scope2/name"
            buddy.log_image("garbage_image", np.zeros((3, 32, 32), dtype=np.float32))
        buddy.log_scope_pop("scope1")


def test_buddy_log_histograms_no_grads(simple_buddy_temporary_data):
    """Check behavior of histogram logging when no gradients exist."""
    model, buddy, data, labels = simple_buddy_temporary_data

    # If we log parameters with no gradients, nothing should happen
    buddy.log_parameters_histogram(ignore_zero_grad=True)

    # If we log gradients with no gradients... throw an error
    with pytest.raises(AssertionError):
        buddy.log_grad_histogram()


def test_buddy_learning_rates(simple_buddy_temporary_data):
    """Check that we can set learning rates. (scalar)"""
    model, buddy, data, labels = simple_buddy_temporary_data

    buddy.set_learning_rate(1e-5)
    assert buddy.get_learning_rate() == 1e-5

    buddy.set_learning_rate(lambda steps: 1e-3)
    assert buddy.get_learning_rate() == 1e-3

    buddy.set_learning_rate(1e-5)
    assert buddy.get_learning_rate() == 1e-5


def test_buddy_learning_rates_lambda(simple_buddy_temporary_data):
    """Check that we can set learning rates. (lambda)"""
    model, buddy, data, labels = simple_buddy_temporary_data

    buddy.set_learning_rate(lambda s: 1e-2)
    assert buddy.get_learning_rate() == 1e-2

    buddy.set_learning_rate(lambda steps: 1e-5)
    assert buddy.get_learning_rate() == 1e-5


def test_buddy_train(simple_buddy_temporary_data):
    """Make sure Buddy losses go down, and that we can write log files without errors."""
    model, buddy, data, labels = simple_buddy_temporary_data
    assert buddy.optimizer_steps == 0
    model.train()

    initial_loss = F.mse_loss(model(data), labels)

    # Try using default learning rate
    buddy.set_default_learning_rate(lambda steps: 1e-3)

    for _ in range(200):
        # Optimize
        loss = F.mse_loss(model(data), labels)
        buddy.minimize(loss)

        # Log for tensorboard
        with buddy.log_scope("scope"):
            buddy.log_scalar("loss", loss)
        buddy.log_parameters_histogram()
        buddy.log_grad_histogram()

    # Flush logs
    buddy.log_writer.flush()

    assert buddy.get_learning_rate() == 1e-3
    assert buddy.optimizer_steps == 200

    # Loss should at least have halved
    final_loss = F.mse_loss(model(data), labels)
    assert final_loss < initial_loss / 2.0
    buddy.save_checkpoint()


def test_buddy_gradient_clipping(simple_buddy_temporary_data):
    """Test that setting the max gradient norm correctly clips the gradients."""

    def _gradient_norm(f: torch.nn.Module) -> float:
        """Compute the gradient norm.

        Args:
            f (torch.nn.Module): Module to check gradients of..

        Returns:
            float: The gradient norm.
        """
        total_norm = 0.0
        for p in f.parameters():
            if p.grad is not None:
                param_norm = p.grad.data.norm(2)
                total_norm += param_norm.item() ** 2
        total_norm = total_norm ** (1.0 / 2)
        return total_norm

    model, buddy, data, labels = simple_buddy_temporary_data
    assert buddy.optimizer_steps == 0
    model.train()

    # Use an extremely small norm value
    max_norm_value = 0.01

    for _ in range(5):
        # Optimize without gradient clipping
        loss = F.mse_loss(model(data), labels)
        buddy.minimize(loss)
        norm_value = _gradient_norm(model)
        assert norm_value > max_norm_value

    for _ in range(5):
        # Optimize with gradient clipping
        loss = F.mse_loss(model(data), labels)
        buddy.minimize(loss, clip_grad_max_norm=max_norm_value)
        norm_value = _gradient_norm(model)

        # All gradients should be clipped
        assert abs(norm_value - max_norm_value) < 1e-5


def test_buddy_train_multiloss_unstable(simple_buddy_temporary_data):
    """Training should be less happy if we (a) use a single optimizer and (b)
    switch abruptly between loss functions with very different scales.
    """
    model, buddy, data, labels = simple_buddy_temporary_data
    assert buddy.optimizer_steps == 0
    model.train()

    initial_loss = F.mse_loss(model(data), labels)
    buddy.set_learning_rate(1e-3)
    for _ in range(50):
        loss = F.mse_loss(model(data), labels) * 1000
        buddy.minimize(loss)
    for _ in range(100):
        loss = F.mse_loss(model(data), labels)
        buddy.minimize(loss)
    for _ in range(50):
        loss = F.mse_loss(model(data), labels) * 1000
        buddy.minimize(loss)

    assert buddy.optimizer_steps == 200

    # Loss will not have halved
    final_loss = F.mse_loss(model(data), labels)
    assert final_loss > initial_loss / 2.0


def test_buddy_train_multiloss_stable(simple_buddy_temporary_data):
    """Training should stabilize if we use separate optimizers for our two
    different loss functions.
    """
    model, buddy, data, labels = simple_buddy_temporary_data
    assert buddy.optimizer_steps == 0
    model.train()

    initial_loss = F.mse_loss(model(data), labels)
    buddy.set_learning_rate(1e-3, "big_loss")
    buddy.set_learning_rate(1e-3, "little_loss")
    for _ in range(50):
        loss = F.mse_loss(model(data), labels) * 1000
        buddy.minimize(loss, optimizer_name="big_loss")
    for _ in range(100):
        loss = F.mse_loss(model(data), labels)
        buddy.minimize(loss, optimizer_name="little_loss")
    for _ in range(50):
        loss = F.mse_loss(model(data), labels) * 1000
        buddy.minimize(loss, optimizer_name="big_loss")

    assert buddy.optimizer_steps == 200

    # Loss should at least have halved
    final_loss = F.mse_loss(model(data), labels)
    assert final_loss < initial_loss / 2.0



import abc
import dataclasses
import pathlib
import signal
import warnings
from glob import escape
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, cast

import dill
import numpy as np
import torch

if TYPE_CHECKING:
    from .._buddy import Buddy
    from ._optimizer import _BuddyOptimizer


class _BuddyCheckpointing(abc.ABC):
    """Buddy's model checkpointing interface."""

    def __init__(self, checkpoint_dir: str, checkpoint_max_to_keep: int) -> None:
        """Checkpointing-specific setup."""
        self._checkpoint_dir = checkpoint_dir
        self._checkpoint_max_to_keep = checkpoint_max_to_keep

        # Find all unlabeled checkpoints for this experiment
        self._checkpoint_unlabeled_files, _ = self._find_checkpoints(
            checkpoint_dir=self._checkpoint_dir,
            experiment_name=cast("Buddy", self)._experiment_name,
            unlabeled_only=True,
        )

    def save_checkpoint(self, label: str = None) -> None:
        """Saves a checkpoint, which can optionally be labeled."""

        checkpoint_dir = pathlib.Path(self._checkpoint_dir)

        # Determine path to checkpoint file
        if label is not None:
            # Label explicitly specified!

            # Numerical labels are reserved for step counts (see below)
            label = cast(str, label)
            assert not label.isdigit()

            experiment_name = cast("Buddy", self)._experiment_name
            path = checkpoint_dir / f"{experiment_name}-{label}.ckpt"
        else:
            # Automatically decide label using global step count
            optimizer_steps = cast("_BuddyOptimizer", self).optimizer_steps
            path = (
                checkpoint_dir
                / f"{cast('Buddy', self)._experiment_name}-{optimizer_steps:016d}.ckpt"
            )

            if (
                self._checkpoint_unlabeled_files
                and path == self._checkpoint_unlabeled_files[-1]
            ):
                cast("Buddy", self)._print("Skipping redundant checkpoint save")
                return

        # Create directory if it doesn't exist yet
        if not checkpoint_dir.exists():
            checkpoint_dir.mkdir(parents=True)

        # Create state to save. This includes:
        # > Model state
        # > Optimizers
        # > Training steps
        # > Buddy configuration
        optimizer_states = {}

        for name, optimizer in cast("_BuddyOptimizer", self)._optimizer_dict.items():
            optimizer_states[name] = optimizer.state_dict()

        state = {
            "optimizer_config": dataclasses.asdict(
                cast("_BuddyOptimizer", self)._optimizer_config
            ),
            "optimizer_states": optimizer_states,
            "state_dict": cast("Buddy", self).model.state_dict(),
        }

        # Ignore SIGINT (eg ctrl+c) events while we save to disk...
        try:
            orig_handler = signal.getsignal(signal.SIGINT)
            signal.signal(signal.SIGINT, lambda _sig, _frame: None)
        except ValueError as e:  # pragma: no cover
            # signal throws a ValueError if we're not in the main thread
            cast("Buddy", self)._print("Error while attaching SIGINT handler:", e)
            orig_handler = None

        # Checkpoint saving
        # > rename is POSIX-compliant and atomic
        tmp_path = checkpoint_dir / f"tmp-{np.random.randint(1e10)}.ckpt"
        torch.save(state, tmp_path, pickle_module=dill)
        tmp_path.rename(path)
        cast("Buddy", self)._print("Saved checkpoint to path:", path)

        # Restore SIGINT handler
        if orig_handler is not None:
            signal.signal(signal.SIGINT, orig_handler)

        # If unlabeled, add to list
        if label is None:
            self._checkpoint_unlabeled_files.append(path)

        # Prune checkpoint files
        while len(self._checkpoint_unlabeled_files) > self._checkpoint_max_to_keep:
            self._checkpoint_unlabeled_files.pop(0).unlink()

    def load_checkpoint_module(
        self,
        source: str,
        target: str = None,
        label: str = None,
        path: str = None,
        experiment_name: str = None,
    ) -> None:
        """Loads parameters from a specific child module within a checkpoint.
        By default, loads the checkpoint with the highest number of training
        iterations.

        Can also be specified via a label or file path.
        """
        assert cast("Buddy", self)._model is not None, "No model attached!"

        if target is None:
            target = source

        # Find and read our checkpoint file
        checkpoint = self._read_checkpoint_file(label, path, experiment_name)

        # Get possible target modules
        module_dict = dict(cast("Buddy", self).model.named_modules())
        assert target in module_dict.keys(), "Nonexistent target module!"

        # Build a state dict for this module only
        source_state_dict = {}
        key_prefix = ""
        if len(source) > 0:
            key_prefix = f"{source}."
        for key, value in checkpoint["state_dict"].items():
            if key.startswith(key_prefix):
                prefix_length = len(key_prefix)
                source_state_dict[key[prefix_length:]] = value

        # Load state dict
        missing, unexpected = module_dict[target].load_state_dict(source_state_dict)
        assert len(missing) == 0
        assert len(unexpected) == 0

        cast("Buddy", self)._print(f"Loaded module: {source} => {target}")

    def load_checkpoint_optimizer(
        self,
        source: str,
        target: str = None,
        label: str = None,
        path: str = None,
        experiment_name: str = None,
    ) -> None:
        """Loads state associated with a specific optimizer from a checkpoint.
        By default, loads the checkpoint with the highest number of training
        iterations.

        Can also be specified via a label or file path.
        """
        assert cast("Buddy", self)._model is not None, "No model attached!"

        if target is None:
            target = source
        cast("_BuddyOptimizer", self)._instantiate_optimizer(target)

        # Find and read our checkpoint file
        checkpoint = self._read_checkpoint_file(label, path, experiment_name)

        # Input validation
        assert (
            source in checkpoint["optimizer_states"].keys()
        ), "Nonexistent source optimizer!"
        assert (
            target in cast("_BuddyOptimizer", self)._optimizer_dict.keys()
        ), "Nonexistent target optimizer!"

        # Load optimizer state
        state_dict = checkpoint["optimizer_states"][source]
        cast("_BuddyOptimizer", self)._optimizer_dict[target].load_state_dict(
            state_dict
        )
        cast("Buddy", self)._print(f"Loaded optimizer: {source} => {target}")

    def load_checkpoint_optimizers(
        self, label=None, path=None, experiment_name=None
    ) -> None:
        """Loads all optimizer settings from a checkpoint. By default, loads
        the checkpoint with the highest number of training iterations.

        Can also be specified via a label or file path.
        """
        assert cast("Buddy", self)._model is not None, "No model attached!"

        # Find and read our checkpoint file
        checkpoint = self._read_checkpoint_file(label, path, experiment_name)

        # Load optimizer state
        self._load_checkpoint_optimizers(checkpoint)

    def load_checkpoint(
        self, label: str = None, path: str = None, experiment_name: str = None
    ) -> None:
        """Loads a checkpoint. By default, loads the one with the highest
        number of training iterations.

        Can also be specified via a label or file path.
        """
        assert cast("Buddy", self)._model is not None, "No model attached!"

        # Find and read our checkpoint file
        checkpoint = self._read_checkpoint_file(label, path, experiment_name)

        # Load optimizer state
        self._load_checkpoint_optimizers(checkpoint)

        # Load model parameters
        cast("Buddy", self).model.load_state_dict(checkpoint["state_dict"], strict=True)

        optimizer_steps = cast("_BuddyOptimizer", self).optimizer_steps
        cast("Buddy", self)._print("Loaded checkpoint at step:", optimizer_steps)
        return

    @property
    def checkpoint_labels(self) -> List[str]:
        """Accessor for listing available checkpoint labels.
        These should be saved as: `experiment_name-label.ckpt` in the
        `checkpoint_dir` directory.

        Returns:
            List[str]: Checkpoint labels, sorted alphabetically.
        """

        experiment_name = cast("Buddy", self)._experiment_name
        checkpoint_dir = pathlib.Path(self._checkpoint_dir)

        # Find all matching checkpoint files
        path_choices = filter(
            lambda path: path.stem.rpartition("-")[0] == experiment_name,
            checkpoint_dir.glob(f"{escape(experiment_name)}-*.ckpt"),
        )

        # Pull out labels
        output = []
        for choice in path_choices:
            label = choice.stem.partition("-")[2]
            output.append(label)

        # Sort output alphabetically and return
        output.sort()
        return output

    def _load_checkpoint_optimizers(self, checkpoint: Dict[str, Any]) -> None:
        # Load Buddy optimizer configuration
        optimizer_config = cast("_BuddyOptimizer", self)._optimizer_config
        for key, value in checkpoint["optimizer_config"].items():
            if hasattr(optimizer_config, key):
                assert type(value) == type(getattr(optimizer_config, key))
                setattr(optimizer_config, key, value)
            else:
                warnings.warn(f"Skipping missing configuration key: {key}={value}")

        # Instantiate optimizers & load state
        for name, state_dict in checkpoint["optimizer_states"].items():
            cast("_BuddyOptimizer", self)._instantiate_optimizer(name)
            cast("_BuddyOptimizer", self)._optimizer_dict[name].load_state_dict(
                state_dict
            )

    def _read_checkpoint_file(
        self,
        label: str = None,
        path: str = None,
        experiment_name: str = None,
        verbose: bool = True,
    ) -> Dict[str, Any]:
        """Find a checkpoint to load.

        This is one of three options:
          1) The latest, based on step # (must have same experiment name)
          2) A file saved with a label (must have same experiment name)
          3) One specified by a path
        """

        # Determine path to checkpoint file
        if path is None and label is None:
            # Load latest unlabeled checkpoint

            # First, find all checkpoint paths
            paths, _ = self._find_checkpoints(
                checkpoint_dir=self._checkpoint_dir,
                experiment_name=cast("Buddy", self)._experiment_name
                if experiment_name is None
                else experiment_name,
            )
            if len(paths) == 0:
                raise FileNotFoundError("Missing checkpoint file")

            # The list of paths will be sorted by optimizer step count
            path = str(paths[-1])

        elif path is None and label is not None:
            # Load a labeled checkpoint
            if experiment_name is None:
                # Use our current experiment name by default
                experiment_name = cast("Buddy", self)._experiment_name
            path = str(
                pathlib.Path(self._checkpoint_dir) / f"{experiment_name}-{label}.ckpt"
            )
        elif path is not None:
            # Load a checkpoint by its location
            path = path
        else:
            assert (
                False
            ), "Too many arguments! Only one of (label, path) is supported at a time."

        # Load checkpoint dict
        checkpoint = torch.load(
            path, map_location=cast("Buddy", self)._device, pickle_module=dill
        )

        # Backwards-compatibility
        # This should eventually be removed :)
        renamed_fields = [
            ("config", "optimizer_config"),
            ("optimizers", "optimizer_states"),
        ]
        for old_name, new_name in renamed_fields:
            if old_name in checkpoint.keys():
                cast("Buddy", self)._print(
                    f"Legacy checkpoint field: {old_name} => {new_name}"
                )
                checkpoint[new_name] = checkpoint[old_name]
                checkpoint.pop(old_name)

        if "steps" in checkpoint.keys():
            cast("Buddy", self)._print("Legacy checkpoint field: steps")
            checkpoint["optimizer_config"]["global_steps"] = checkpoint["steps"]
            checkpoint.pop("steps")

        # Checkpoint file validation
        valid_keys = set(["optimizer_config", "optimizer_states", "state_dict"])
        for key in checkpoint.keys():
            assert key in valid_keys

        # Raise warning for optimizer type mismatches
        if (
            hasattr(self, "_optimizer_config")
            and checkpoint["optimizer_config"]["optimizer_type"]
            != cast("_BuddyOptimizer", self)._optimizer_config.optimizer_type
        ):
            warnings.warn("Checkpoint loading: overriding optimizer type.")

        if verbose:
            cast("Buddy", self)._print("Read checkpoint from path:", path)
        return checkpoint

    def _find_checkpoints(
        self, checkpoint_dir: str, experiment_name: str, unlabeled_only: bool = False
    ) -> Tuple[List[pathlib.Path], Dict[pathlib.Path, int]]:
        """(Private) Returns a list of all unlabeled checkpoints associated
        with this experiment, sorted from oldest to newest.
        """

        # Find all matching checkpoint files
        path_choices = filter(
            lambda path: path.stem.rpartition("-")[0] == experiment_name,
            pathlib.Path(checkpoint_dir).glob(f"{escape(experiment_name)}-*.ckpt"),
        )

        # Find checkpoint files + associated step counts
        paths = []
        step_counts = {}
        for choice in path_choices:
            steps = None
            string_steps = choice.stem.partition("-")[2]
            try:
                # Try unlabeled checkpoint
                step_counts[choice] = int(string_steps)
                paths.append(choice)
            except ValueError:
                # If label can't be cast to integer: this is a labeled checkpoint
                # Only incle unlabeled if unlabeled_only is False
                if not unlabeled_only:
                    paths.append(choice)

        # Add step counts for labeled checkpoints
        for path in paths:
            if path in step_counts:
                continue
            if unlabeled_only:
                # This condition should never be hit
                assert False

            steps = self._read_checkpoint_file(path=str(path), verbose=False)[
                "optimizer_config"
            ]["global_steps"]
            step_counts[path] = steps

        # Sort output by steps
        paths.sort(key=lambda path: step_counts[path])

        # Return paths & steps
        return paths, step_counts



from typing import Dict

import numpy as np
import torch.nn.functional as F

import fannypack

from ..fixtures import simple_buddy_temporary_data, simple_net


def test_freeze_integration(simple_buddy_temporary_data):
    """Integration test for module freezing + minimizing with Buddy."""
    model, buddy, data, labels = simple_buddy_temporary_data

    # Minimize loss
    loss = F.mse_loss(model(data), labels)
    buddy.minimize(loss)

    # Get NumPy representation of parameters, then copy
    layer4_params_before: Dict[str, np.ndarray] = fannypack.utils.SliceWrapper(
        fannypack.utils.to_numpy(dict(model.layer4.named_parameters()))
    ).map(np.array)

    print(layer4_params_before)

    # Freeze layer4, then minimize
    fannypack.utils.freeze_module(model.layer4)
    for _ in range(20):
        loss = F.mse_loss(model(data), labels)
        buddy.minimize(loss)

    layer4_params_after: Dict[str, np.ndarray] = fannypack.utils.to_numpy(
        dict(model.layer4.named_parameters())
    )

    # Compare (frozen) layer4 parameters before/after second minimize
    assert set(layer4_params_before.keys()) == set(layer4_params_after.keys())
    for key in layer4_params_after.keys():
        np.testing.assert_allclose(layer4_params_before[key], layer4_params_after[key])


def test_freeze_unfreeze(simple_net):
    """Check to make sure freezing/unfreezing a module behaves as expected."""

    # All parameters should be initially unfrozen
    _check_frozen(simple_net, False)

    # Freeze the module
    fannypack.utils.freeze_module(simple_net)

    # All parameters should be frozen
    _check_frozen(simple_net, True)

    # Unfreeze the module
    fannypack.utils.unfreeze_module(simple_net)

    # All parameters should be unfrozen
    _check_frozen(simple_net, False)


def test_unfreeze_children(simple_net):
    """Check to make sure unfreezing child modules behaves as expected."""

    # All parameters should be initially unfrozen
    _check_frozen(simple_net, False)

    # Freeze the module
    fannypack.utils.freeze_module(simple_net)

    # All parameters should be frozen
    _check_frozen(simple_net, True)

    # Unfreeze the module
    fannypack.utils.unfreeze_module(simple_net.layer1)
    fannypack.utils.unfreeze_module(simple_net.layer2)
    fannypack.utils.unfreeze_module(simple_net.layer3)
    fannypack.utils.unfreeze_module(simple_net.layer4)

    # All parameters should be unfrozen
    _check_frozen(simple_net, False)


def test_freeze_children(simple_net):
    """Check to make sure freezing child modules behaves as expected."""

    # All parameters should be initially unfrozen
    _check_frozen(simple_net, False)

    # Freeze a child module
    fannypack.utils.freeze_module(simple_net.layer1)

    # All parameters for layer1 should be frozen
    _check_frozen(simple_net.layer1, True)
    _check_frozen(simple_net.layer2, False)
    _check_frozen(simple_net.layer3, False)
    _check_frozen(simple_net.layer4, False)

    # Unfreeze the entire module
    fannypack.utils.unfreeze_module(simple_net)

    # All parameters should be unfrozen
    _check_frozen(simple_net, False)


# Helper to check whether a module is frozen or not
def _check_frozen(module, is_frozen):
    for param in module.parameters():
        assert param.requires_grad == (not is_frozen)



import functools
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import torch

# Note that we define two container TypeVars:
# - A bound container, whose output type exactly matches the input type
#       -> For moving across devices, for example, a List[torch.Tensor] will always
#          return a List[torch.Tensor]
# - And a constrained container, whose output type is only guaranteed to match at the
#   top-level
#       -> For converting NumPy -> Torch, a List[np.ndarray] will return a
#          List[torch.Tensor]; this is really difficult to represent in Python, so we
#          just mark the output as a List[Any]

BoundContainer = TypeVar("BoundContainer", bound=Union[Dict, List, Tuple])
Key = TypeVar("Key")

_ConstrainedContainer = TypeVar("_ConstrainedContainer", Dict, List, Tuple)
_InputType = TypeVar("_InputType", np.ndarray, torch.Tensor)
_OutputType = TypeVar("_OutputType", np.ndarray, torch.Tensor)


@overload
def _convert_recursive(
    x: _InputType,
    convert: Callable[[_InputType], _OutputType],
    input_type: Type,
) -> _OutputType:
    ...


@overload
def _convert_recursive(
    x: _ConstrainedContainer,
    convert: Callable[[_InputType], _OutputType],
    input_type: Type,
) -> _ConstrainedContainer:
    ...


def _convert_recursive(x, convert, input_type):
    """Private conversion helper. Recursively calls a conversion function on inputs
    within a nested set of containers.
    """

    # Conversion base case
    if isinstance(x, input_type):
        x = cast(_InputType, x)
        return convert(x)

    # Convert containers: bind arguments to helper function
    convert_recursive = functools.partial(
        _convert_recursive, convert=convert, input_type=input_type
    )

    # Convert dictionaries of values
    if isinstance(x, dict):
        x = cast(dict, x)
        return dict(zip(x.keys(), map(convert_recursive, x.values())))

    # Convert lists of values
    if isinstance(x, list):
        x = cast(list, x)
        return list(map(convert_recursive, x))

    # Convert tuples of values
    if isinstance(x, tuple):
        x = cast(tuple, x)
        if hasattr(x, "_fields"):  # NamedTuple
            return type(x)(*map(convert_recursive, x))
        else:
            return tuple(map(convert_recursive, x))

    # Unsupported input types
    assert False, f"Unsupported datatype {type(x)}!"


@overload
def to_device(
    x: torch.Tensor, device: torch.device, detach: bool = False
) -> torch.Tensor:
    ...


@overload
def to_device(
    x: BoundContainer, device: torch.device, detach: bool = False
) -> BoundContainer:
    ...


def to_device(x, device, detach=False):
    """Move a torch tensor, list, tuple (standard or named), or dict of tensors to a
    different device. Recursively applied for nested containers.

    Args:
        x (torch.Tensor, list, tuple (standard or named), or dict): Tensor or container
            of tensors to move.
        device (torch.device): Target device.
        detach (bool, optional): If set, detaches tensors after moving. Defaults to
            False.

    Returns:
        torch.Tensor, list, tuple (standard or named), or dict: Output, type will mirror
        input.
    """

    def convert(x):
        if detach:
            x = x.detach()
        return x.to(device)

    return _convert_recursive(x, convert=convert, input_type=torch.Tensor)


# This a lot of boilerplate, but seems impossible to reproduce same level of specificity
# with TypeVars


@overload
def to_torch(
    x: np.ndarray,
    device: str = "cpu",
    convert_doubles_to_floats: bool = True,
) -> torch.Tensor:
    ...


@overload
def to_torch(
    x: List[np.ndarray],
    device: str = "cpu",
    convert_doubles_to_floats: bool = True,
) -> List[torch.Tensor]:
    ...


@overload
def to_torch(
    x: List,
    device: str = "cpu",
    convert_doubles_to_floats: bool = True,
) -> List:
    ...


@overload
def to_torch(
    x: Tuple[np.ndarray, ...],
    device: str = "cpu",
    convert_doubles_to_floats: bool = True,
) -> Tuple[torch.Tensor, ...]:
    ...


@overload
def to_torch(
    x: Tuple,
    device: str = "cpu",
    convert_doubles_to_floats: bool = True,
) -> Tuple:
    ...


@overload
def to_torch(
    x: Dict[Key, np.ndarray],
    device: str = "cpu",
    convert_doubles_to_floats: bool = True,
) -> Dict[Key, torch.Tensor]:
    ...


@overload
def to_torch(
    x: Dict[Key, Any],
    device: str = "cpu",
    convert_doubles_to_floats: bool = True,
) -> Dict[Key, Any]:
    ...


def to_torch(
    x,
    device="cpu",
    convert_doubles_to_floats=True,
):
    """Converts a NumPy array, list, tuple (standard or named), or dict of NumPy arrays
    for use in PyTorch.  Recursively applied for nested containers.

    Args:
        x (np.ndarray, list, tuple (standard or named), or dict): Array or container of
            arrays to convert to torch tensors.
        device (torch.device, optional): Torch device to create tensors on. Defaults to
            `"cpu"`.
        convert_doubles_to_floats (bool, optional): If set, converts 64-bit floats to
            32-bit. Defaults to True.

    Returns:
        torch.Tensor, list, tuple (standard or named), or dict: Output, type will mirror input.
    """

    def convert(x: np.ndarray) -> torch.Tensor:
        output = torch.from_numpy(x)
        if x.dtype == np.float64 and convert_doubles_to_floats:
            output = output.float()
        output = output.to(device)
        return output

    return _convert_recursive(x, convert=convert, input_type=np.ndarray)


@overload
def to_numpy(x: torch.Tensor) -> np.ndarray:
    ...


@overload
def to_numpy(x: List[torch.Tensor]) -> List[np.ndarray]:
    ...


@overload
def to_numpy(x: List) -> List:
    ...


@overload
def to_numpy(x: Tuple[torch.Tensor, ...]) -> Tuple[np.ndarray, ...]:
    ...


@overload
def to_numpy(x: Tuple) -> Tuple:
    ...


@overload
def to_numpy(x: Dict[Key, torch.Tensor]) -> Dict[Key, np.ndarray]:
    ...


@overload
def to_numpy(x: Dict[Key, Any]) -> Dict[Key, Any]:
    ...


def to_numpy(x):
    """Converts a tensor, list, tuple (standard or named), or dict of tensors for use in
    Numpy. Recursively applied for nested containers.

    Args:
        x (torch.Tensor, list, tuple (standard or named), or dict): Tensor or container
            of tensors to convert to NumPy.

    Returns:
        np.ndarray, list, tuple (standard or named), or dict: Output, type will mirror input.
    """

    def convert(x: torch.Tensor) -> np.ndarray:
        output = x.detach().cpu().numpy()
        return output

    return _convert_recursive(x, convert=convert, input_type=torch.Tensor)



import abc
import argparse
from typing import List, Type

from ._utils import BuddyPaths

subcommand_registry: List[Type["Subcommand"]] = []


class Subcommand(abc.ABC):
    """Subcommand interface: defines arguments, runtime routine."""

    subcommand: str
    helptext: str

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """Registers a subcommand."""
        super().__init_subclass__(**kwargs)
        subcommand_registry.append(cls)

    @classmethod
    @abc.abstractmethod
    def add_arguments(
        cls, *, parser: argparse.ArgumentParser, paths: BuddyPaths
    ) -> None:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def main(cls, *, args: argparse.Namespace, paths: BuddyPaths) -> None:
        raise NotImplementedError



from typing import TYPE_CHECKING
import platform

if TYPE_CHECKING or platform.python_version().startswith("3.6."):
    from . import data, nn, utils
else:
    # Lazy submodule loading
    def __getattr__(name):
        import importlib

        module = importlib.import_module(__name__)
        if name not in __all__:
            raise AttributeError(f"{__name__!r} has no attribute {name!r}")
        imported = importlib.import_module(f".{name}", module.__spec__.parent)
        setattr(module, name, imported)
        return imported

__all__ = ["data", "nn", "utils"]



from typing import Any, Dict, NamedTuple, Tuple

import numpy as np
import pytest
import torch

import fannypack


@pytest.fixture
def nested_conversion() -> Tuple[Dict, Dict, Dict]:
    X = {"data": ([np.zeros((10, 1, 5, 1, 10))], np.eye(3))}
    X_torch = fannypack.utils.to_torch(X)
    X_numpy = fannypack.utils.to_numpy(X_torch)
    return X, X_torch, X_numpy


def test_not_same(nested_conversion: Tuple[Dict, Dict, Dict]):
    """Make sure we've received three distinct dictionaries."""
    X, X_torch, X_numpy = nested_conversion
    assert X is not X_torch
    assert X is not X_numpy


def test_conversion_shapes(nested_conversion: Tuple[Dict, Dict, Dict]):
    """Check numpy <=> torch conversions."""
    X, X_torch, X_numpy = nested_conversion
    assert X_torch["data"][0][0].shape == X_numpy["data"][0][0].shape
    assert X_torch["data"][1].shape == X_numpy["data"][1].shape


def test_conversion_types(nested_conversion: Tuple[Dict, Dict, Dict]):
    """Check numpy <=> torch conversions."""
    X, X_torch, X_numpy = nested_conversion
    assert type(X_numpy["data"][0][0]) == np.ndarray
    assert type(X_torch["data"][0][0]) == torch.Tensor
    assert type(X_numpy["data"][1]) == np.ndarray
    assert type(X_torch["data"][1]) == torch.Tensor


def test_conversion_values(nested_conversion: Tuple[Dict, Dict, Dict]):
    """Check numpy <=> torch conversions."""
    X, X_torch, X_numpy = nested_conversion
    assert np.allclose(X["data"][0][0], X_numpy["data"][0][0])
    assert np.allclose(X["data"][1], X_numpy["data"][1])


def test_conversion_failure():
    """Smoke test for an unsupported input type."""
    with pytest.raises(AssertionError):
        fannypack.utils.to_torch(None)


def test_to_device():
    """Smoke test for to_device."""
    X = {"data": [np.zeros((10, 1, 5, 1, 10))]}
    X_torch = fannypack.utils.to_torch(X)

    X_new = fannypack.utils.to_device(X_torch, torch.device("cpu"), detach=True)
    assert X_torch["data"][0].shape == X_new["data"][0].shape
    assert type(X_new["data"][0]) == torch.Tensor


def test_named_tuple():
    """Check that we can convert named tuples."""

    class P(NamedTuple):
        x: Any
        y: Any

    p = P(np.array(1), np.array(2))
    p_torch = fannypack.utils.to_torch(p)
    assert type(p_torch.x) == torch.Tensor
    assert type(p_torch.y) == torch.Tensor
    assert int(p_torch.x) == 1
    assert int(p_torch.y) == 2



import argparse
import datetime
import os

import beautifultable
import termcolor

from ._subcommand import Subcommand
from ._utils import BuddyPaths, format_size, get_size


class ListSubcommand(Subcommand):
    """Get & summarize existing Buddy experiments."""

    subcommand: str = "list"
    table_column_headers = [
        "Name",
        "Ckpts",
        "Logs",
        "Meta",
        "Timestamp",
    ]

    @classmethod
    def add_arguments(
        cls, *, parser: argparse.ArgumentParser, paths: BuddyPaths
    ) -> None:
        parser.add_argument(
            "--sort-by",
            type=str,
            choices=[h.lower() for h in cls.table_column_headers[:-1]],
            help="Sort experiment table by column. Defaults to timestamp.",
        )

    @classmethod
    def main(cls, *, args: argparse.Namespace, paths: BuddyPaths) -> None:
        results = paths.find_experiments(verbose=True)

        # Generate dynamic-width table
        try:
            terminal_columns = int(os.popen("stty size", "r").read().split()[1])
        except IndexError:
            # stty size fails when run from outside proper terminal (eg in tests)
            terminal_columns = 100
        table = beautifultable.BeautifulTable(maxwidth=min(100, terminal_columns))
        table.set_style(beautifultable.STYLE_BOX_ROUNDED)
        table.rows.separator = ""

        # Add bolded headers
        table.columns.header = [
            termcolor.colored(h, attrs=["bold"]) for h in cls.table_column_headers
        ]

        # Constant for "not applicable" fields
        NA = termcolor.colored("N/A", "red")

        # Add experiment rows, oldest to newest
        sorted_experiment_names = sorted(
            results.experiment_names,
            key=lambda n: results.timestamps.get(n, 0.0),
        )
        for name in sorted_experiment_names:
            # Get checkpoint count
            checkpoint_count = 0
            if name in results.checkpoint_counts:
                checkpoint_count = results.checkpoint_counts[name]

            # Get timestamp
            timestamp = ""
            if name in results.timestamps:
                timestamp = datetime.datetime.fromtimestamp(
                    results.timestamps[name]
                ).strftime(
                    "%b %d, %y @ %-H:%M" if terminal_columns > 100 else "%y-%m-%d"
                )

            # Add row for experiment
            table.rows.append(
                [
                    name,
                    checkpoint_count,
                    termcolor.colored(
                        format_size(get_size(paths.get_log_dir(name)), short=True),
                        "green",
                    )
                    if name in results.log_experiments
                    else NA,
                    termcolor.colored(
                        format_size(
                            get_size(paths.get_metadata_file(name)), short=True
                        ),
                        "green",
                    )
                    if name in results.metadata_experiments
                    else NA,
                    timestamp,
                ]
            )

        # Print table, sorted by name
        print(f"Found {len(results.experiment_names)} experiments!")
        if args.sort_by is not None:
            # Sort-by field: lowercase -> index
            table.sort(
                list(map(lambda s: s.lower(), cls.table_column_headers)).index(
                    args.sort_by
                )
            )
        print(table)



import warnings
from typing import Callable


def deprecation_wrapper(message: str, function_or_class: Callable) -> Callable:
    """Creates a wrapper for a deprecated function or class. Prints a warning
    the first time a function or class is called.

    Args:
        message (str): Warning message.
        function_or_class (Callable): Function or class to wrap.

    Returns:
        Callable: Wrapped function/class.
    """

    warned = False

    def curried(*args, **kwargs):  # pragma: no cover
        nonlocal warned
        if not warned:
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            warned = True
        return function_or_class(*args, **kwargs)

    return curried


def new_name_wrapper(
    old_name: str, new_name: str, function_or_class: Callable
) -> Callable:
    """Creates a wrapper for a renamed function or class. Prints a warning the first
    time a function or class is called with the old name.

    Args:
        old_name (str): Old name of function or class. Printed in warning.
        new_name (str): New name of function or class. Printed in warning.
        function_or_class (Callable): Function or class to wrap.

    Returns:
        Callable: Wrapped function/class.
    """
    return deprecation_wrapper(
        f"{old_name} is deprecated! Use {new_name} instead.", function_or_class
    )



import os
import signal

import fannypack


class FakePDB:
    def __init__(self):
        self.called = False

    def set_trace(self):
        self.called = True


def test_sigint():
    """Check that we open PDB when a SIGINT event is handled."""
    fake_pdb = FakePDB()

    fannypack.utils.pdb_safety_net()
    fannypack.utils._pdb_safety_net.pdb = fake_pdb

    assert not fake_pdb.called
    os.kill(os.getpid(), signal.SIGINT)
    assert fake_pdb.called



from typing import Optional, Union

import numpy as np
import torch


def cholesky_inverse(u: torch.Tensor, upper: bool = False) -> torch.Tensor:
    """Alternative to `torch.cholesky_inverse()`, with support for batch dimensions.

    Relevant issue tracker: https://github.com/pytorch/pytorch/issues/7500

    Args:
        u (torch.Tensor): Triangular Cholesky factor. Shape should be `(*, N, N)`.
        upper (bool, optional): Whether to consider the Cholesky factor as a lower or
            upper triangular matrix.

    Returns:
        torch.Tensor:
    """
    if u.dim() == 2 and not u.requires_grad:
        return torch.cholesky_inverse(u, upper=upper)
    return torch.cholesky_solve(torch.eye(u.size(-1)).expand(u.size()), u, upper=upper)


def cholupdate(
    L: torch.Tensor,
    x: torch.Tensor,
    weight: Optional[Union[torch.Tensor, float]] = None,
) -> torch.Tensor:
    """Batched rank-1 Cholesky update.

    Computes the Cholesky decomposition of `RR^T + weight * xx^T`.

    Args:
        L (torch.Tensor): Lower triangular Cholesky decomposition of a PSD matrix. Shape
            should be `(*, matrix_dim, matrix_dim)`.
        x (torch.Tensor): Rank-1 update vector. Shape should be `(*, matrix_dim)`.
        weight (torch.Tensor or float, optional): Set to -1 for "downdate". Shape must
            be broadcastable with `(*, matrix_dim)`.

    Returns:
        torch.Tensor: New L matrix. Same shape as L.
    """
    # Expected shapes: (*, dim, dim) and (*, dim)
    batch_dims = L.shape[:-2]
    matrix_dim = x.shape[-1]
    assert x.shape[:-1] == batch_dims
    assert matrix_dim == L.shape[-1] == L.shape[-2]

    # Flatten batch dimensions, and clone for tensors we need to mutate
    L = L.reshape((-1, matrix_dim, matrix_dim))
    x = x.reshape((-1, matrix_dim)).clone()
    L_out_cols = []

    sign: Union[float, torch.Tensor]
    if weight is None:
        sign = L.new_ones((1,))
    elif isinstance(weight, float):
        x = x * np.sqrt(np.abs(weight))
        sign = float(np.sign(weight))
    else:
        x = x * torch.sqrt(torch.abs(weight))
        sign = torch.sign(weight)

    # Cholesky update; mostly copied from Wikipedia:
    # https://en.wikipedia.org/wiki/Cholesky_decomposition
    for k in range(matrix_dim):
        r = torch.sqrt(L[:, k, k] ** 2 + sign * x[:, k] ** 2)
        c = (r / L[:, k, k])[:, None]
        s = (x[:, k] / L[:, k, k])[:, None]

        # We build output column-by-column to avoid in-place modification errors
        L_out_col = torch.zeros_like(L[:, :, k])
        L_out_col[:, k] = r
        L_out_col[:, k + 1 :] = (L[:, k + 1 :, k] + sign * s * x[:, k + 1 :]) / c
        L_out_cols.append(L_out_col)

        # We clone x at each iteration, also to avoid in-place modification errors
        x_next = x.clone()
        x_next[:, k + 1 :] = c * x[:, k + 1 :] - s * L_out_col[:, k + 1 :]
        x = x_next

    # Stack columns together
    L_out = torch.stack(L_out_cols, dim=2)

    # Unflatten batch dimensions and return
    return L_out.reshape(batch_dims + (matrix_dim, matrix_dim))


def quadratic_matmul(x: torch.Tensor, A: torch.Tensor) -> torch.Tensor:
    r"""Computes $x^\top A x$, with support for arbitrary batch axes.

    Stolen from @alberthli/@wuphilipp.

    Args:
        x (torch.Tensor): Vectors. Shape should be `(*, D)`.
        A (torch.Tensor): Matrices. Shape should be `(*, D, D)`.

    Returns:
        torch.Tensor: Batched output of multiplication. Shape should be `(*)`.
    """
    assert x.shape[-1] == A.shape[-1] == A.shape[-2]

    x_T = x.unsqueeze(-2)  # shape=(*, 1, X)
    x_ = x.unsqueeze(-1)  # shape(*, X, 1)
    quadratic = x_T @ A @ x_  # shape=(*, 1, 1)
    return quadratic.squeeze(-1).squeeze(-1)


def gaussian_log_prob(
    mean: torch.Tensor, covariance: torch.Tensor, value: torch.Tensor
) -> torch.Tensor:
    """Computes log probabilities under multivariate Gaussian distributions,
    with support for arbitrary batch axes.

    Naive version of...
    ```
    torch.distributions.MultivariateNormal(
        mean, covariance
    ).log_prob(value)
    ```
    that avoids some Cholesky-related CUDA errors.
    https://discuss.pytorch.org/t/cuda-illegal-memory-access-when-using-batched-torch-cholesky/51624

    Stolen from @alberthli/@wuphilipp.

    Args:
        mean (torch.Tensor): Means vectors. Shape should be `(*, D)`.
        covariance (torch.Tensor): Covariances matrices. Shape should be `(*, D, D)`.
        value (torch.Tensor): State vectors. Shape should be `(*, D)`.

    Returns:
        torch.Tensor: Batched log probabilities. Shape should be `(*)`.
    """
    D = mean.shape[-1]
    assert covariance.shape[:-1] == mean.shape == value.shape

    exponential = quadratic_matmul(value - mean, torch.inverse(covariance))
    other_terms = torch.logdet(covariance) + D * np.log(2.0 * np.pi)
    log_p = -0.5 * (exponential + other_terms)
    return log_p


def matrix_dim_from_tril_count(tril_count: int):
    """Computes the dimension of a lower triangular square matrix given a count of its
    lower-triangular components.

    Args:
        tril_count (int): Count of lower-triangular terms.
    Returns:
        int: Dimension of square matrix.
    """
    matrix_dim = int(0.5 * (1 + 8 * tril_count) ** 0.5 - 0.5)
    return matrix_dim


def tril_count_from_matrix_dim(matrix_dim: int):
    """Computes the number of lower triangular terms in a square matrix of a given
    dimension `(matrix_dim, matrix_dim)`.

    Args:
        matrix_dim (int): Dimension of square matrix.
    Returns:
        int: Count of lower-triangular terms.
    """
    tril_count = (matrix_dim ** 2 - matrix_dim) // 2 + matrix_dim
    return tril_count


def tril_from_vector(lower_vector: torch.Tensor) -> torch.Tensor:
    """Computes lower-triangular square matrices from a flattened vector of nonzero
    terms. Supports arbitrary batch dimensions.

    Args:
        lower_vector (torch.Tensor): Vectors containing the nonzero terms of a
            square lower-triangular matrix. Shape should be `(*, tril_count)`.
    Returns:
        torch.Tensor: Square matrices. Shape should be `(*, matrix_dim, matrix_dim)`.
    """
    batch_dims = lower_vector.shape[:-1]
    tril_count = lower_vector.shape[-1]
    matrix_dim = matrix_dim_from_tril_count(tril_count)

    output = torch.zeros(batch_dims + (matrix_dim, matrix_dim))
    tril_indices = torch.tril_indices(matrix_dim, matrix_dim)
    output[..., tril_indices[0], tril_indices[1]] = lower_vector
    return output


def vector_from_tril(tril_matrix: torch.Tensor) -> torch.Tensor:
    """Retrieves the lower triangular terms of square matrices as vectors. Supports
    arbitrary batch dimensions.

    Args:
        tril_matrix (torch.Tensor): Square matrices. Shape should be
            `(*, matrix_dim, matrix_dim)`
    Returns:
        torch.Tensor: Flattened vectors. Shape should be `(*, tril_count)`.
    """
    matrix_dim = tril_matrix.shape[-1]
    assert tril_matrix.shape[-2] == matrix_dim

    tril_indices = torch.tril_indices(matrix_dim, matrix_dim)
    return tril_matrix[..., tril_indices[0], tril_indices[1]]


def tril_inverse(tril_matrix: torch.Tensor) -> torch.Tensor:
    """Invert a lower-triangular matrix.

    Args:
        tril_matrix (torch.Tensor): Lower-triangular matrix to invert. Shape should be
            `(*, matrix_dim, matrix_dim)`.

    Returns:
        torch.Tensor: Inverted matrix. Shape should be `(*, matrix_dim, matrix_dim)`.
    """
    assert tril_matrix.shape[-1] == tril_matrix.shape[-2], "Input must be square!"
    matrix_dim = tril_matrix.shape[-1]
    batch_dims = tril_matrix.shape[:-2]

    # Get an identity matrix
    identity = torch.eye(tril_matrix.shape[-1], device=tril_matrix.device)

    # View identity w/ batch dimensions
    identity = identity.reshape(
        (1,) * len(batch_dims) + (matrix_dim, matrix_dim)
    ).expand(tril_matrix.shape)

    # Invert with triangular solve
    inverse = torch.triangular_solve(identity, tril_matrix, upper=False).solution
    assert inverse.shape == tril_matrix.shape
    return inverse



import argparse
import os

import beautifultable
import termcolor
from pygments import formatters, highlight, lexers

import fannypack

from ._subcommand import Subcommand
from ._utils import BuddyPaths, format_size, get_size


class InfoSubcommand(Subcommand):
    """Print info about a Buddy experiment: checkpoints, metadata, etc."""

    subcommand: str = "info"

    @classmethod
    def add_arguments(
        cls, *, parser: argparse.ArgumentParser, paths: BuddyPaths
    ) -> None:
        parser.add_argument(
            "experiment_name",
            type=str,
            help="Name of experiment, as printed by `$ buddy list`.",
            metavar="experiment_name",  # Set metavar => don't show choices in help menu
            choices=paths.find_experiments().experiment_names,
        )

    @classmethod
    def main(cls, *, args: argparse.Namespace, paths: BuddyPaths) -> None:
        # Get experiment name
        experiment_name = args.experiment_name
        print(experiment_name)

        # Generate dynamic-width table
        try:
            terminal_columns = int(os.popen("stty size", "r").read().split()[1])
        except IndexError:
            # stty size fails when run from outside proper terminal (eg in tests)
            terminal_columns = 100
        table = beautifultable.BeautifulTable(
            maxwidth=min(100, terminal_columns),
            default_alignment=beautifultable.ALIGN_LEFT,
        )
        table.set_style(beautifultable.STYLE_BOX_ROUNDED)

        def add_table_row(label, value):
            table.rows.append([termcolor.colored(label, attrs=["bold"]), value])

        # Constant for "not applicable" fields
        NA = termcolor.colored("N/A", "red")

        # Find checkpoint files
        checkpoint_paths = paths.find_checkpoints(experiment_name)

        # Display size, labels of checkpoints
        if len(checkpoint_paths) > 0:
            checkpoint_total_size = 0
            checkpoint_labels = []
            buddy = fannypack.utils.Buddy(experiment_name, verbose=False)
            checkpoint_paths, steps = buddy._find_checkpoints(
                paths.checkpoint_dir, args.experiment_name
            )
            for checkpoint_path in checkpoint_paths:
                prefix = os.path.join(paths.checkpoint_dir, f"{experiment_name}-")
                suffix = ".ckpt"
                assert str(checkpoint_path).startswith(prefix)
                assert str(checkpoint_path).endswith(suffix)
                label = str(checkpoint_path)[len(prefix) : -len(suffix)]

                checkpoint_labels.append(f"{label} (steps: {steps[checkpoint_path]})")
                checkpoint_total_size += get_size(str(checkpoint_path))

            add_table_row(
                "Total checkpoint size",
                termcolor.colored(format_size(checkpoint_total_size), "green"),
            )
            add_table_row(
                "Average checkpoint size",
                termcolor.colored(
                    format_size(checkpoint_total_size / len(checkpoint_paths)), "green"
                ),
            )
            add_table_row("Checkpoint labels", "\n".join(checkpoint_labels))
        else:
            add_table_row("Total checkpoint size", NA)
            add_table_row("Average checkpoint size", NA)
            add_table_row("Checkpoint labels", "")

        # Display log file size
        log_path = paths.get_log_dir(args.experiment_name)
        if os.path.exists(log_path):
            #  _delete(log_path, args.forever)
            add_table_row(
                "Log size", termcolor.colored(format_size(get_size(log_path)), "green")
            )
        else:
            add_table_row("Log size", NA)

        # Display metadata + metadata size
        metadata_path = paths.get_metadata_file(args.experiment_name)
        if os.path.exists(metadata_path):
            add_table_row(
                "Metadata size",
                termcolor.colored(format_size(get_size(metadata_path)), "green"),
            )
            with open(metadata_path, "r") as f:
                add_table_row(
                    "Metadata",
                    highlight(
                        f.read().strip(),
                        lexers.YamlLexer(),
                        formatters.TerminalFormatter(),
                    ),
                )
        else:
            add_table_row("Metadata size", NA)
            add_table_row("Metadata", NA)

        # Print table
        print(table)



from ..data import _trajectories_file
from ._buddy import Buddy
from ._conversions import to_device, to_numpy, to_torch
from ._deprecation import deprecation_wrapper, new_name_wrapper
from ._git import get_git_commit_hash
from ._math import (
    cholesky_inverse,
    cholupdate,
    gaussian_log_prob,
    matrix_dim_from_tril_count,
    quadratic_matmul,
    tril_count_from_matrix_dim,
    tril_from_vector,
    tril_inverse,
    vector_from_tril,
)
from ._module_freezing import freeze_module, unfreeze_module
from ._pdb_safety_net import pdb_safety_net
from ._slice_wrapper import SliceWrapper
from ._squeeze import squeeze

DictIterator = new_name_wrapper(
    "fannypack.utils.DictIterator", "fannypack.utils.SliceWrapper", SliceWrapper
)

TrajectoriesFile = new_name_wrapper(
    "fannypack.utils.TrajectoriesFiles",
    "fannypack.data.TrajectoriesFile",
    _trajectories_file.TrajectoriesFile,
)

__all__ = [
    "Buddy",
    "to_device",
    "to_numpy",
    "to_torch",
    "deprecation_wrapper",
    "new_name_wrapper",
    "get_git_commit_hash",
    "cholesky_inverse",
    "cholupdate",
    "gaussian_log_prob",
    "matrix_dim_from_tril_count",
    "quadratic_matmul",
    "tril_count_from_matrix_dim",
    "tril_from_vector",
    "tril_inverse",
    "vector_from_tril",
    "freeze_module",
    "unfreeze_module",
    "pdb_safety_net",
    "SliceWrapper",
    "squeeze",
]



from typing import Dict

import torch.nn as nn

_freeze_restore_values: Dict[nn.Module, Dict] = {}


def freeze_module(module: nn.Module, recurse: bool = True) -> None:
    """Freeze the weights of a PyTorch module by setting the `requires_grad` attributes
    of enclosed parameters to `False`.

    Args:
        module (torch.nn.Module): Module to freeze.
        recurse (bool, optional): If True, then recursively freezes children.
            Otherwise, only freezes immediate parameters.
    """

    global _freeze_restore_values

    # Recursively call on children
    if recurse:
        for child in module.children():
            freeze_module(child)

    # Do nothing if module is already frozen
    if module in _freeze_restore_values:
        return

    # Freeze parameters
    restore_values = {}
    for name, parameter in module.named_parameters(recurse=False):
        restore_values[name] = parameter.requires_grad
        parameter.requires_grad = False

        # We need to reset the gradient value to None: otherwise, the gradient will just
        # be set to zero when we call zero_grad(), and the optimizer will still update
        # it via momentum, weight decay, etc.
        parameter.grad = None

    _freeze_restore_values[module] = restore_values


def unfreeze_module(module: nn.Module, recurse: bool = True) -> None:
    """Unfreeze the weights of a PyTorch module, which needs to have been
    frozen with :func:`fannypack.utils.freeze_module`. Restores all original values
    `requires_grad` values.

    Args:
        module (torch.nn.Module): Module to unfreeze.
        recurse (bool, optional): If True, then recursively unfreezes children.
            Otherwise, only unfreezes immediate parameters.
    """

    global _freeze_restore_values

    # Recursively call on children
    if recurse:
        for child in module.children():
            unfreeze_module(child)

    # Do nothing if module is already unfrozen
    if module not in _freeze_restore_values:
        return

    # Freeze parameters
    restore_values = _freeze_restore_values.pop(module)
    for name, parameter in module.named_parameters(recurse=False):
        parameter.requires_grad = restore_values[name]



import argparse
import dataclasses
import os
import shutil

import simple_term_menu
import termcolor
from pygments import formatters, highlight, lexers

from ._subcommand import Subcommand
from ._utils import BuddyPaths

_TRASH_DIR = "./_trash/"


class DeleteSubcommand(Subcommand):
    """Delete one or more Buddy experiments.

    If no experiment names are passed in, displays a menu.
    """

    subcommand: str = "delete"

    @classmethod
    def add_arguments(
        cls, *, parser: argparse.ArgumentParser, paths: BuddyPaths
    ) -> None:
        parser.add_argument(
            "experiment_name",
            type=str,
            help="Name of experiment, as printed by `$ buddy list`.",
            metavar="experiment_name",  # Set metavar => don't show choices in help menu
            choices=paths.find_experiments().experiment_names | {""},
            nargs="*",
            default="",  # Need default to prevent unhashable list error
        )
        parser.add_argument(
            "--forever",
            action="store_true",
            help=f"Delete experiment forever: if unset, move files into `{_TRASH_DIR}`.",
        )

    @classmethod
    def main(cls, *, args: argparse.Namespace, paths: BuddyPaths) -> None:
        # Get experiment name
        experiment_names = args.experiment_name

        if len(experiment_names) > 0:
            # Delete specified experiments
            for name in experiment_names:
                try:
                    cls._delete_experiment(name, args.forever, paths)
                except RuntimeError as e:
                    print(f"Encountered error for {name}")
                    print(e)
        else: #  pragma: no cover
            print("Navigate: j/k \t Select: <CR>")
            print(
                termcolor.colored(
                    "There will be no confirmation message.", "red", attrs=["bold"]
                )
            )
            while True:
                delete_options = sorted(paths.find_experiments().experiment_names)
                if len(delete_options) == 0:
                    print("No experiments to delete!")
                    return

                # Show menu
                menu = simple_term_menu.TerminalMenu(
                    delete_options,
                    preview_command=lambda name: cls.get_pretty_metadata(name, paths),
                    preview_size=0.75,
                )
                experiment_name = menu.show()

                # Quit on Ctrl+C, Esc, etc
                if experiment_name is None:
                    return

                # Delete experiments
                cls._delete_experiment(
                    delete_options[experiment_name], args.forever, paths
                )
                paths.clear_cache()

    @classmethod
    def _delete_experiment(
        cls, experiment_name: str, forever: bool, paths: BuddyPaths
    ) -> None:
        # If we're just moving an experiment, check that it doesn't exist already
        if not forever:
            new_checkpoint_files = dataclasses.replace(
                paths, checkpoint_dir=os.path.join(_TRASH_DIR, paths.checkpoint_dir)
            ).find_checkpoints(experiment_name)
            if len(new_checkpoint_files) != 0:
                raise RuntimeError(
                    "Checkpoints for matching experiment name already exist in trash; "
                    "rename experiment before deleting."
                )
            if os.path.exists(
                os.path.join(_TRASH_DIR, paths.log_dir, f"{experiment_name}")
            ):
                raise RuntimeError(
                    "Logs for matching experiment name already exist in trash; "
                    "rename experiment before deleting."
                )
            if os.path.exists(
                os.path.join(_TRASH_DIR, paths.metadata_dir, f"{experiment_name}.yaml")
            ):
                raise RuntimeError(
                    "Metadata for matching experiment name already exist in trash; "
                    "rename experiment before deleting."
                )

        # Delete checkpoint files
        checkpoint_paths = paths.find_checkpoints(experiment_name)
        print(f"Found {len(checkpoint_paths)} checkpoint files")
        for path in checkpoint_paths:
            cls._delete(path, forever)

        # Delete metadata
        metadata_path = os.path.join(paths.metadata_dir, f"{experiment_name}.yaml")
        if os.path.exists(metadata_path):
            cls._delete(metadata_path, forever)
        else:
            print("No metadata found")

        # Delete logs
        log_path = os.path.join(paths.log_dir, f"{experiment_name}")
        if os.path.exists(log_path):
            cls._delete(log_path, forever)
        else:
            print("No logs found")

    @staticmethod
    def _delete(path: str, forever: bool) -> None:
        assert os.path.exists(path)

        if forever:
            # Delete file/directory forever
            if os.path.isdir(path):
                print(f"Deleting {path} (recursive)")
                shutil.rmtree(path)
            elif os.path.isfile(path):
                print(f"Deleting {path}")
                os.remove(path)
            else:
                assert False, "Something went wrong"
        else:
            # Move files/directory to a new path
            new_path = os.path.join(_TRASH_DIR, path)

            # Create trash directory if it doesn't exist yet
            directory = os.path.dirname(new_path)
            if not os.path.isdir(directory):
                os.makedirs(directory)

            # Move file/directory to trash
            print(f"Moving {path} to {new_path}")
            os.rename(path, new_path)

    @staticmethod
    def get_pretty_metadata(experiment_name: str, paths: BuddyPaths) -> str:
        try:
            with open(paths.get_metadata_file(experiment_name), "r") as f:
                return highlight(
                    f.read().strip(),
                    lexers.YamlLexer(),
                    formatters.TerminalFormatter(),
                )
        except FileNotFoundError:
            return termcolor.colored("No metadata", "blue")



import abc
import contextlib
import pathlib
from typing import TYPE_CHECKING, Generator, List, Optional, Union, cast

import numpy as np
import torch.utils.tensorboard

from .. import _deprecation

if TYPE_CHECKING:
    from .._buddy import Buddy
    from ._optimizer import _BuddyOptimizer


class _BuddyLogging(abc.ABC):
    """Buddy's TensorBoard logging interface."""

    def __init__(self, log_dir: str) -> None:
        """Logging-specific setup.

        Args:
            log_dir (str): Path to save Tensorboard logs to.
        """
        self._log_dir = log_dir

        # Backwards-compatibility for deprecated API
        self.log = _deprecation.new_name_wrapper(
            "Buddy.log()", "Buddy.log_scalar()", self.log_scalar
        )
        self.log_model_grad_hist = _deprecation.new_name_wrapper(
            "Buddy.log_model_grad_hist()",
            "Buddy.log_grad_histogram()",
            self.log_grad_histogram,
        )
        self.log_model_weight_hist = _deprecation.new_name_wrapper(
            "Buddy.log_model_weight_hist()",
            "Buddy.log_parameters_histogram()",
            self.log_parameters_histogram,
        )

        # State variables for TensorBoard
        # Note that the writer is lazily instantiated; see below
        self._log_writer: Optional[torch.utils.tensorboard.SummaryWriter] = None
        self._log_scopes: List[str] = []

    @contextlib.contextmanager
    def log_scope(self, scope: str) -> Generator[None, None, None]:
        """Returns a context manager that scopes log names.

        Example usage:

        ```
            with buddy.log_scope("scope"):
                # Logs to scope/loss
                buddy.log_scalar("loss", loss_tensor)
        ```

        Args:
            scope (str): Name of scope.
        """
        self.log_scope_push(scope)
        yield
        self.log_scope_pop(scope)

    def log_scope_push(self, scope: str) -> None:
        """Push a scope to log tensors into.

        Example usage:
        ```
            buddy.log_scope_push("scope")

            # Logs to scope/loss
            buddy.log_scalar("loss", loss_tensor)

            buddy.log_scope_pop("scope") # name parameter is optional

        Args:
            scope (str): Name of scope.
        ```
        """
        self._log_scopes.append(scope)

    def log_scope_pop(self, scope: str = None) -> None:
        """Pop a scope we logged tensors into. See `log_scope_push()`.

        Args:
            scope (str, optional): Name of scope. Needs to be the top one in the stack.
        """
        popped = self._log_scopes.pop()
        if scope is not None:
            assert popped == scope, f"{popped} does not match {scope}!"

    def log_image(
        self,
        name: str,
        image: Union[torch.Tensor, np.ndarray],
        dataformats: str = "CHW",
    ) -> None:
        """Convenience function for logging an image tensor for visualization in
        TensorBoard.

        Equivalent to:
        ```
        buddy.log_writer.add_image(
            buddy.log_scope_prefix(name),
            image,
            buddy.optimizer_steps,
            dataformats
        )
        ```

        Args:
            name (str): Identifier for Tensorboard.
            image (torch.Tensor or np.ndarray): Image to log.
            dataformats (str, optional): Dimension ordering. Defaults to "CHW".
        """
        # Add scope prefixes
        name = self.log_scope_prefix(name)

        # Log scalar
        optimizer_steps = cast("_BuddyOptimizer", self).optimizer_steps
        self.log_writer.add_image(
            name, image, global_step=optimizer_steps, dataformats=dataformats
        )

    def log_scalar(
        self, name: str, value: Union[torch.Tensor, np.ndarray, float]
    ) -> None:
        """Convenience function for logging a scalar for visualization in TensorBoard.

        Equivalent to:
        ```
        buddy.log_writer.add_scalar(
            buddy.log_scope_prefix(name),
            value,
            buddy.optimizer_steps
        )
        ```

        Args:
            name (str): Identifier for Tensorboard.
            value (torch.Tensor, np.ndarray, or float): Value to log.
        """
        # Add scope prefixes
        name = self.log_scope_prefix(name)

        # Log scalar
        optimizer_steps = cast("_BuddyOptimizer", self).optimizer_steps
        self.log_writer.add_scalar(name, value, global_step=optimizer_steps)

    def log_parameters_histogram(
        self, scope: str = "weights", *, ignore_zero_grad: bool = True
    ) -> None:
        """Log model weights into a histogram.

        Naming: with `scope` set to "weights", a parameter name "model.Linear.bias" will be
        logged to the tag `buddy.log_scope_prefix("weights/model/Linear/bias")`.

        Args:
            scope (str, optional): Scope to log gradients into. Defaults to "weights".
            ignore_zero_grad (bool, optional): Ignore parameters without gradients:
                decreases log sizes when only parts of models are being trained.
                Defaults to True.
        """
        optimizer_steps = cast("_BuddyOptimizer", self).optimizer_steps

        with self.log_scope(scope):
            for param_name, p in cast("Buddy", self).model.named_parameters():
                if ignore_zero_grad and p.grad is None:
                    continue

                param_name = param_name.replace(".", "/")
                self.log_writer.add_histogram(
                    tag=self.log_scope_prefix(param_name),
                    values=p.data.detach().cpu().numpy(),
                    global_step=optimizer_steps,
                )

    def log_grad_histogram(self, scope: str = "grad") -> None:
        """Log model gradients into a histogram. Should be called after
        `buddy.minimize()`.

        Naming: with `scope` set to "grad", a parameter name "model.Linear.bias" will be
        logged to the tag `buddy.log_scope_prefix("grad/model/Linear/bias")`.

        Args:
            scope (str, optional): Scope to log gradients into. Defaults to "grad".
        """
        optimizer_steps = cast("_BuddyOptimizer", self).optimizer_steps

        found_gradients = False
        with self.log_scope(scope):
            for param_name, p in cast("Buddy", self).model.named_parameters():
                if p.grad is None:
                    continue
                param_name = param_name.replace(".", "/")
                self.log_writer.add_histogram(
                    tag=self.log_scope_prefix(param_name),
                    values=p.grad.detach().cpu().numpy(),
                    global_step=optimizer_steps,
                )
                found_gradients = True

        assert found_gradients, "No gradients found!"

    def log_scope_prefix(self, name: str = "") -> str:
        """Get or apply the current log scope prefix.

        Example usage:
        ```
        print(buddy.log_scope_prefix()) # ""

        with buddy.log_scope("scope0"):
            print(buddy.log_scope_prefix("loss")) # "scope0/loss"

            with buddy.log_scope("scope1"):
                print(buddy.log_scope_prefix()) # "scope0/scope1/"
        ```

        Args:
            name (str, optional): Name to prepend a prefix to. Defaults to an empty string.

        Returns:
            str: Scoped log name, or scope prefix if input is empty.
        """
        if len(self._log_scopes) == 0:
            return name
        return "{}/{}".format("/".join(self._log_scopes), name)

    @property
    def log_writer(self) -> torch.utils.tensorboard.SummaryWriter:
        """Accessor for standard Tensorboard SummaryWriter. Instantiated lazily."""
        if self._log_writer is None:
            self._log_writer = torch.utils.tensorboard.SummaryWriter(
                pathlib.Path(self._log_dir) / cast("Buddy", self)._experiment_name
            )
        return self._log_writer



import fannypack.utils
import numpy as np
import torch


def test_cholesky_inverse():
    """Checks that our Cholesky inverse matches `torch.cholesky_inverse()`."""
    torch.autograd.set_detect_anomaly(True)
    batch_dims = (5,)
    matrix_dim = 3
    L = fannypack.utils.tril_from_vector(
        torch.randn(
            batch_dims + (fannypack.utils.tril_count_from_matrix_dim(matrix_dim),),
        )
    )

    for i, our_inverse in enumerate(fannypack.utils.cholesky_inverse(L)):
        torch.testing.assert_allclose(our_inverse, torch.cholesky_inverse(L[i]))


def test_cholesky_inverse_no_batch():
    """Checks that our Cholesky inverse matches `torch.cholesky_inverse()`."""
    torch.autograd.set_detect_anomaly(True)
    matrix_dim = 3
    L = fannypack.utils.tril_from_vector(
        torch.randn(fannypack.utils.tril_count_from_matrix_dim(matrix_dim))
    )
    torch.testing.assert_allclose(
        fannypack.utils.cholesky_inverse(L), torch.cholesky_inverse(L)
    )


def test_cholesky_inverse_backward():
    """Checks backward pass support for Cholesky inverses."""
    batch_dims = (5,)
    matrix_dim = 3
    L = fannypack.utils.tril_from_vector(
        torch.randn(
            batch_dims + (fannypack.utils.tril_count_from_matrix_dim(matrix_dim),),
            requires_grad=True,
        )
    )
    torch.sum(fannypack.utils.cholesky_inverse(L)).backward()


def test_cholupdate():
    """Checks rank-1 Cholesky update forward pass."""
    batch_dims = (1,)  # (5, 3, 7)
    matrix_dim = 2
    L = fannypack.utils.tril_from_vector(
        torch.randn(
            batch_dims + (fannypack.utils.tril_count_from_matrix_dim(matrix_dim),)
        )
    )
    x = torch.randn(batch_dims + (matrix_dim,))

    updated_L = fannypack.utils.cholupdate(L, x)

    torch.testing.assert_allclose(
        L @ L.transpose(-1, -2) + x[..., :, None] @ x[..., None, :],
        updated_L @ updated_L.transpose(-1, -2),
    )


def test_cholupdate_negative():
    """Checks rank-1 Cholesky update forward pass, with weight set to -2."""
    batch_dims = tuple()
    matrix_dim = 3
    L = fannypack.utils.tril_from_vector(
        torch.randn(
            batch_dims + (fannypack.utils.tril_count_from_matrix_dim(matrix_dim),)
        )
    ) + torch.eye(matrix_dim)
    x = torch.ones(matrix_dim) * 0.2

    # Make sure our output will be PSD
    L = torch.cholesky(L @ L.transpose(-1, -2) + x[..., :, None] @ x[..., None, :])

    updated_L = fannypack.utils.cholupdate(
        L, x, weight=torch.tensor(-0.5, dtype=torch.float32)
    )

    torch.testing.assert_allclose(
        L @ L.transpose(-1, -2) - 0.5 * x[..., :, None] @ x[..., None, :],
        updated_L @ updated_L.transpose(-1, -2),
    )


def test_cholupdate_negative_float_weight():
    """Checks rank-1 Cholesky update forward pass, with weight set to -2."""
    batch_dims = tuple()
    matrix_dim = 3
    L = fannypack.utils.tril_from_vector(
        torch.randn(
            batch_dims + (fannypack.utils.tril_count_from_matrix_dim(matrix_dim),)
        )
    ) + torch.eye(matrix_dim)
    x = torch.ones(matrix_dim) * 0.2

    # Make sure our output will be PSD
    L = torch.cholesky(L @ L.transpose(-1, -2) + x[..., :, None] @ x[..., None, :])

    updated_L = fannypack.utils.cholupdate(L, x, weight=-0.5)

    torch.testing.assert_allclose(
        L @ L.transpose(-1, -2) - 0.5 * x[..., :, None] @ x[..., None, :],
        updated_L @ updated_L.transpose(-1, -2),
    )


def test_cholupdate_backward():
    """Smoke test for rank-1 Cholesky update backward pass."""
    torch.autograd.set_detect_anomaly(True)
    batch_dims = (5, 3, 7)
    matrix_dim = 5
    L = fannypack.utils.tril_from_vector(
        torch.randn(
            batch_dims + (fannypack.utils.tril_count_from_matrix_dim(matrix_dim),),
            requires_grad=True,
        )
    )
    x = torch.randn(batch_dims + (matrix_dim,))

    updated_L = fannypack.utils.cholupdate(L, x)

    # If the Cholesky update is implemented incorrectly, we'll get an "inplace
    # operation" RuntimeError here
    torch.sum(updated_L).backward()


def test_quadratic_matmul():
    """Tests quadratic_matmul() by checking its backward pass."""
    N = 100
    D = 5

    # Deterministic test
    torch.random.manual_seed(0)

    # Compute quadratic
    x = torch.randn((N, D), requires_grad=True)
    A = torch.randn((N, D, D))
    xTAx = fannypack.utils.quadratic_matmul(x, A)

    # Check shape
    assert xTAx.shape == (N,)

    # Check gradient
    torch.sum(xTAx).backward()
    np.testing.assert_allclose(
        *fannypack.utils.to_numpy(
            [x.grad, ((A.transpose(-1, -2) + A) @ x[:, :, None]).squeeze()]
        ),
        rtol=1e-4,
        atol=1e-6,
    )


def test_gaussian_log_prob():
    """Check that our Gaussian log probability implementation matches the native PyTorch
    one.
    """
    N = 100
    D = 5

    # Deterministic test
    torch.random.manual_seed(0)

    # Compute quadratic
    mean = torch.randn((N, D))
    value = torch.randn((N, D))
    covariance = torch.randn((N, D, D))
    covariance = covariance @ covariance.transpose(-1, -2)

    # Check output
    np.testing.assert_allclose(
        *fannypack.utils.to_numpy(
            [
                torch.distributions.MultivariateNormal(
                    loc=mean, covariance_matrix=covariance
                ).log_prob(value=value),
                fannypack.utils.gaussian_log_prob(
                    mean=mean,
                    covariance=covariance,
                    value=value,
                ),
            ]
        ),
        rtol=1e-3,
        atol=1e10,
    )


def test_tril_count_simple():
    """Basic counting utility check."""
    assert fannypack.utils.matrix_dim_from_tril_count(6) == 3
    assert fannypack.utils.tril_count_from_matrix_dim(3) == 6


def test_tril_count_bijective():
    """Check bijectivity of our counting utilities."""

    for matrix_dim in range(1, 100):
        assert (
            fannypack.utils.matrix_dim_from_tril_count(
                fannypack.utils.tril_count_from_matrix_dim(matrix_dim)
            )
            == matrix_dim
        )


def test_tril_vector_conversions():
    """Check that our `tril_to_vector` and `vector_to_tril` functions return the correct
    shapes + bijectivity.
    """

    for batch_dims in ((), (3, 2, 5), (7,)):
        for matrix_dim in range(1, 10):
            tril_count = fannypack.utils.tril_count_from_matrix_dim(matrix_dim)
            vectors = torch.randn(batch_dims + (tril_count,))
            trils = fannypack.utils.tril_from_vector(vectors)

            assert trils.shape == batch_dims + (matrix_dim, matrix_dim)
            assert torch.all(fannypack.utils.vector_from_tril(trils) == vectors)


def test_tril_inverse():
    """Check that our `tril_inverse` function correctly inverts some full-rank
    matrices.
    """
    for matrix_dim in range(2, 5):
        tril_matrix = fannypack.utils.tril_from_vector(
            torch.randn((5, fannypack.utils.tril_count_from_matrix_dim(matrix_dim)))
        )
        inverse = fannypack.utils.tril_inverse(tril_matrix)

        torch.testing.assert_allclose(tril_matrix @ inverse, torch.eye(matrix_dim))



from ._drive_files import cached_drive_file, download_drive_file, set_cache_path
from ._trajectories_file import TrajectoriesFile

__all__ = [
    "cached_drive_file",
    "download_drive_file",
    "set_cache_path",
    "TrajectoriesFile",
]



from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import torch

import fannypack

# Valid raw types that we can wrap
_raw_types = set([list, tuple, np.ndarray, torch.Tensor])

# Generic types
MapOutputType = TypeVar("MapOutputType")
WrappedType = TypeVar(
    "WrappedType",
    bound=Union[
        List,
        Tuple,
        torch.Tensor,
        np.ndarray,
        Dict[Any, List],
        Dict[Any, Tuple],
        Dict[Any, torch.Tensor],
        Dict[Any, np.ndarray],
    ],
)


class SliceWrapper(Iterable, Generic[WrappedType]):
    """A wrapper class for creating a unified interface for slicing and manipulating:
    - Lists
    - Tuples
    - Torch tensors
    - Numpy arrays
    - Dictionaries containing a same-length group of any of the above

    This makes it easy to read, slice, and index into blocks of data organized into
    dictionaries.

    Nominally:
    ```
    dataset = SliceWrapper({
        "features": features,
        "labels": labels,
    })

    train_count = 100
    train_dataset = dataset[:train_count]
    val_dataset = dataset[train_count:]
    ```
    would be equivalent to:
    ```
    train_count = 100
    train_dataset = {
        "features": features[:train_count],
        "labels": b[:train_count],
    }
    val_dataset = {
        "features": features[train_count:],
        "labels": b[train_count:],
    }
    ```

    For convenience, a transparent interface is provided for iterables that are directly
    wrapped. Thus:
    ```
    SliceWrapper([1, 2, 3])[::-1]
    ```
    would return:
    ```
    [1, 2, 3][::-1]
    ```
    """

    def __init__(self, data: WrappedType):
        self.data: WrappedType = data
        """list, tuple, torch.Tensor, np.ndarray, or dict: Wrapped data."""

        # Sanity checks
        if type(self.data) == dict:
            # Cast for type-checking
            data_dict = cast(dict, self.data)

            # Every value in the dict should have the same length & type
            content_length = None
            content_type = None
            for value in data_dict.values():
                assert content_length is None or len(value) == content_length
                assert content_type is None or type(value) == content_type
                content_length = len(value)
                content_type = type(value)
                assert content_type in _raw_types
        else:
            # Non-dictionary inputs
            assert type(data) in _raw_types, "Unsupported datatype!"

        # Backwards-compatibility
        def convert_to_numpy():  # pragma: no cover
            self.data.update(self.map(np.asarray))

        self.convert_to_numpy = fannypack.utils.deprecation_wrapper(
            "SliceWrapper.convert_to_numpy() is deprecated -- please use "
            "the functional SliceWrapper.map() interface instead!",
            convert_to_numpy,
        )

    def __getitem__(self, index: Any) -> Any:
        """Unified interface for indexing into our wrapped object; shorthand for
        `SliceWrapper.map(lambda v: v[index])`.

        For wrapped dictionaries, this returns a new (un-wrapped) dictionary with the
        index applied value-wise.
        Thus:
        ```
        SliceWrapper({
            "a": a,
            "b": b,
        })[index]
        ```
        would return:
        ```
        {
            "a": a[index],
            "b": b[index],
        }
        ```

        For iterables that are directly wrapped, this is equivalent to evaluating
        `data[index]`.
        Thus..
        ```
        SliceWrapper([1, 2, 3])[::-1]
        ```
        would return:
        ```
        [1, 2, 3][::-1]
        ```

        Args:
            index (Any): Index. Can be a slice, tuple, boolean array, etc.

        Returns:
            Any: Indexed value. See overall function docstring.
        """
        return cast(WrappedType, self.map(lambda v: v[index]))

    def __len__(self) -> int:
        """Unified interface for evaluating the length of a wrapped object.

        Equivalent to `SliceWrapper.shape[0]`.

        Returns:
            int: Length of wrapped object.
        """
        return self.shape[0]

    def __iter__(self):
        """Iterable __iter__() interface."""
        self._iter_index = 0
        return self

    def __next__(self):
        """Iterable __next__() interface."""
        try:
            output = self[self._iter_index]
            self._iter_index += 1
            return output
        except IndexError:
            pass
        raise StopIteration

    def append(self, other: Any) -> None:
        """Append to the end of our data object.

        Only supported for wrapped lists and dictionaries containing lists.

        For wrapped lists, this is equivalent to `data.append(other)`.

        For dictionaries, `other` should be a dictionary.
        Behavior example:
        ```
        # Data before append
        {"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}

        # Value of other
        {"a": 5, "b": 3}

        # Data after append
        {"a": [1, 2, 3, 4, 5], "b": [5, 6, 7, 8, 3]}
        ```

        Args:
            other (Any): Object to append.
        """
        if type(self.data) == dict:
            assert type(other) == dict, "Appended object must be a dictionary"

            # Cast for type-checking
            data_dict = cast(dict, self.data)
            other_dict = cast(dict, other)

            for key, value in other_dict.items():
                if key in data_dict.keys():
                    assert (
                        type(data_dict[key]) == list
                    ), "Append is only supported for wrapped lists"
                    data_dict[key].append(value)
                else:
                    data_dict[key] = [value]
        elif type(self.data) is list:
            cast(List, self.data).append(other)
        else:
            assert False, "Append is only supported for wrapped lists"

    def extend(self, other: WrappedType) -> None:
        """Extend to the end of our data object.

        Only supported for wrapped lists and dictionaries containing lists.

        For wrapped lists, this is equivalent to `data.extend(other)`.

        For dictionaries, `other` should be a dictionary.
        Behavior example:
        ```
        # Data before extend
        {"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]}

        # Value of other
        {"a": [5], "b": [3]}

        # Data after extend
        {"a": [1, 2, 3, 4, 5], "b": [5, 6, 7, 8, 3]}
        ```

        Args:
            other (dict or list): Object to append.
        """
        if type(self.data) == dict:
            assert type(other) == dict

            # Cast for type-checking
            data_dict = cast(dict, self.data)
            other_dict = cast(dict, other)

            for key, value in other_dict.items():
                if key in data_dict.keys():
                    assert (
                        type(data_dict[key]) == list
                    ), "Extend is only supported for wrapped lists"
                    data_dict[key].extend(value)
                else:
                    data_dict[key] = value
        elif type(self.data) == list:
            cast(list, self.data).extend(other)
        else:
            assert False, "Extend is only supported for wrapped lists"

    @overload
    def map(
        self: Union[
            "SliceWrapper[Dict[Any, List]]",
            "SliceWrapper[Dict[Any, Tuple]]",
            "SliceWrapper[Dict[Any, torch.Tensor]]",
            "SliceWrapper[Dict[Any, np.ndarray]]",
        ],
        function: Callable[[Any], MapOutputType],
    ) -> Dict[Any, MapOutputType]:
        pass

    @overload
    def map(
        self: Union[
            "SliceWrapper[List]",
            "SliceWrapper[Tuple]",
            "SliceWrapper[torch.Tensor]",
            "SliceWrapper[np.ndarray]",
        ],
        function: Callable[[Any], MapOutputType],
    ) -> MapOutputType:
        pass

    @overload
    def map(
        self, function: Callable[[Any], MapOutputType]
    ) -> Union[MapOutputType, Dict[Any, MapOutputType]]:
        pass

    def map(self, function):
        """Apply a function to all iterables within our wrapped data object.

        For iterables that are directly wrapped (eg lists), this is equivalent to
        evaluating:
        ```
        slice_wrapper: SliceWrapper[List]
        function(slice_wrapper.data)
        ```

        For dictionaries, `function` is applied value-wise.
        Thus, an input of:
        ```
        SliceWrapper({
            "a": [1, 2, 3],
            "b": [2, 4, 5],
        })
        ```
        would return:
        ```
        {
            "a": function([1, 2, 3]),
            "b": function([2, 4, 5]),
        }
        ```

        Args:
            function (Callable): Function to map.
        """
        if type(self.data) == dict:
            # Cast for type-checking
            data_dict = cast(dict, self.data)

            # Construct output
            mapped_data = {}
            for key, value in data_dict.items():
                mapped_data[key] = function(value)
            return mapped_data
        elif type(self.data) in _raw_types:
            return function(self.data)
        else:
            assert False, f"Unsupported data type: {type(self.data)}"

    @property
    def shape(self) -> Tuple[int, ...]:
        """Unified interface for polling the shape of our wrapped object.

        For lists and tuples, this evaluates to `(len(data),)`.

        For Numpy arrays and torch tensors, we get `data.shape`.

        For dictionaries, we return a tuple containing all shared dimensions between
        our wrapped values, starting from the leftmost dimension.

        Args:

        Returns:
            Tuple[int, ...]:
        """
        if type(self.data) == dict:
            # Cast for type-checking
            data_dict = cast(dict, self.data)

            # Find longest shared shape prefix
            output: Tuple[int, ...]
            first = True
            for value in data_dict.values():
                shape = self._shape_helper(value)
                if first:
                    output = shape
                    first = False
                    continue

                for i in range(min(len(output), len(shape))):
                    if output[i] != shape[i]:
                        output = output[:i]
                        break

            return tuple(output)
        elif type(self.data) in _raw_types:
            return self._shape_helper(self.data)
        else:
            assert False, "Unsupported datatype!"

    @staticmethod
    def _shape_helper(data) -> Tuple[int, ...]:
        """Computes the shape of an object. `data.shape` for tensors and Numpy arrays,
        `(len(data))` for lists and tuples.
        """
        if type(data) in (torch.Tensor, np.ndarray):
            # Return full shape
            return data.shape
        elif type(data) in (list, tuple):
            # Return 1D shape
            return (len(data),)
        else:
            assert False, "Unsupported datatype!"



import argparse
import glob
import os

from ._subcommand import Subcommand
from ._utils import BuddyPaths


class RenameSubcommand(Subcommand):
    """Rename a Buddy experiment."""

    subcommand: str = "rename"

    @classmethod
    def add_arguments(
        cls, *, parser: argparse.ArgumentParser, paths: BuddyPaths
    ) -> None:
        parser.add_argument(
            "source",
            type=str,
            help="Current name of experiment, as printed by `$ buddy list`.",
            metavar="source",  # Set metavar => don't show choices in help menu
            choices=paths.find_experiments().experiment_names,
        )
        parser.add_argument("dest", type=str, help="New name of experiment.")

    @classmethod
    def main(cls, *, args: argparse.Namespace, paths: BuddyPaths) -> None:
        # Get old, new experiment names
        old_experiment_name = args.source
        new_experiment_name = args.dest

        # Validate that new experiment name doesn't exist
        new_checkpoint_files = glob.glob(
            os.path.join(
                paths.checkpoint_dir, f"{glob.escape(new_experiment_name)}-*.ckpt"
            )
        )
        if len(new_checkpoint_files) != 0:
            raise RuntimeError(
                f"Checkpoints already exist for destination name: {new_experiment_name}"
            )
        if os.path.exists(paths.get_log_dir(new_experiment_name)):
            raise RuntimeError(
                f"Logs already exist for destination name: {new_experiment_name}"
            )
        if os.path.exists(paths.get_metadata_file(new_experiment_name)):
            raise RuntimeError(
                f"Metadata already exist for destination name: {new_experiment_name}"
            )

        # Move checkpoint files
        checkpoint_paths = paths.find_checkpoints(old_experiment_name)
        print(f"Found {len(checkpoint_paths)} checkpoint files")
        for path in checkpoint_paths:
            # Get new checkpoint path
            prefix = os.path.join(paths.checkpoint_dir, f"{old_experiment_name}-")
            suffix = ".ckpt"
            assert path.startswith(prefix)
            assert path.endswith(suffix)
            label = path[len(prefix) : -len(suffix)]
            new_path = os.path.join(
                paths.checkpoint_dir, f"{new_experiment_name}-{label}.ckpt"
            )

            # Move checkpoint
            print(f"> Moving {path} to {new_path}")
            os.rename(path, new_path)

        # Move metadata
        metadata_path = paths.get_metadata_file(old_experiment_name)
        if os.path.exists(metadata_path):
            new_path = paths.get_metadata_file(new_experiment_name)
            print(f"Moving {metadata_path} to {new_path}")
            os.rename(metadata_path, new_path)
        else:
            print("No metadata found")

        # Move logs
        log_path = paths.get_log_dir(old_experiment_name)
        if os.path.exists(log_path):
            new_path = paths.get_log_dir(new_experiment_name)
            print(f"Moving {log_path} to {new_path}")
            os.rename(log_path, new_path)
        else:
            print("No logs found")
